# Copyright (C) 2011-2014 Red Hat, Inc.
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; If not, see <http://www.gnu.org/licenses/>.
#
# Author: tasleson
import os
import sys
from lsm import (Volume, NfsExport, Capabilities, Pool, System, Battery,
                 Disk, AccessGroup, FileSystem, FsSnapshot,
                 uri_parse, LsmError, ErrorNumber,
                 INetworkAttachedStorage, TargetPort)

from lsm._common import return_requires as _return_requires
from lsm._common import UDS_PATH as _UDS_PATH
from lsm._transport import TransPort as _TransPort
from lsm._data import IData as _IData

import six


# Removes self for the hash d
# @param    d   Hash to remove self from
# @returns d with hash removed.
def _del_self(d):
    """
    Used to remove the self key from the dict d.  Self is included when calling
    the function locals() in a class method.
    """
    del d['self']
    return d


def _check_search_key(search_key, supported_keys):
    if search_key and search_key not in supported_keys:
        raise LsmError(ErrorNumber.UNSUPPORTED_SEARCH_KEY,
                       "Unsupported search_key: '%s'" % search_key)
    return


# Descriptive exception about daemon not running.
def _raise_no_daemon():
    raise LsmError(ErrorNumber.DAEMON_NOT_RUNNING,
                   "The libStorageMgmt daemon is not running (process "
                   "name lsmd), try 'service libstoragemgmt start'")


# Main client class for library.
# ** IMPORTANT **
# Theory of operation for methods in this class.
# We are using the name of the method and the name of the parameters and
# using python introspection abilities to translate them to the method and
# parameter names.  Makes the code compact, but you will break things if the
# IPlugin class does not match the method names and parameters here!
class Client(INetworkAttachedStorage):

    #
    # Used for default flag value
    #
    FLAG_RSVD = 0

    FLAG_VOLUME_CREATE_USE_SYSTEM_CACHE = 1 << 0
    FLAG_VOLUME_CREATE_USE_IO_PASSTHROUGH = 1 << 1
    FLAG_VOLUME_CREATE_DISABLE_SYSTEM_CACHE = 1 << 2
    FLAG_VOLUME_CREATE_DISABLE_IO_PASSTHROUGH = 1 << 3

    """
    Client side class used for managing storage that utilises RPC mechanism.
    """
    # Method added so that the interface for the client RPC and the plug-in
    # itself match.
    def plugin_register(self, uri, plain_text_password, timeout_ms, flags=0):
        raise RuntimeError("Do not call directly!")

    # Called when we are ready to initialize the plug-in.
    # @param    self                    The this pointer
    # @param    uri                     The uniform resource identifier
    # @param    plain_text_password     Password as plain text
    # @param    timeout_ms              The timeout in ms
    # @param    flags                   Reserved for future use, must be zero.
    # @returns None
    def __start(self, uri, password, timeout, flags=0):
        """
        Instruct the plug-in to get ready
        """
        self._tp.rpc('plugin_register', _del_self(locals()))

    # Checks to see if any unix domain sockets exist in the base directory
    # and opens a socket to one to see if the server is actually there.
    # @param    self    The this pointer
    # @returns True if daemon appears to be present, else false.
    @staticmethod
    def _check_daemon_exists():
        uds_path = Client._plugin_uds_path()
        if os.path.exists(uds_path):
            for root, sub_folders, files in os.walk(uds_path):
                for filename in files:
                    uds = os.path.join(root, filename)

                    try:
                        # This operation will work if the daemon is available
                        s = _TransPort.get_socket(uds)
                        s.close()
                        return True
                    except LsmError:
                        pass
        else:
            # Base directory is not present?
            pass
        return False

    @staticmethod
    def _plugin_uds_path():
        rc = _UDS_PATH

        if 'LSM_UDS_PATH' in os.environ:
            rc = os.environ['LSM_UDS_PATH']

        return rc

    # Class constructor
    # @param    self                    The this pointer
    # @param    uri                     The uniform resource identifier
    # @param    plain_text_password     Password as plain text (Optional)
    # @param    timeout_ms              The timeout in ms
    # @param    flags                   Reserved for future use, must be zero.
    # @returns None
    def __init__(self, uri, plain_text_password=None, timeout_ms=30000,
                 flags=0):
        self._uri = uri
        self._password = plain_text_password
        self._timeout = timeout_ms
        self._uds_path = Client._plugin_uds_path()

        u = uri_parse(uri, ['scheme'])

        scheme = u['scheme']
        if "+" in scheme:
            (plug, proto) = scheme.split("+")
            scheme = plug

        self.plugin_path = os.path.join(self._uds_path, scheme)

        if os.path.exists(self.plugin_path):
            self._tp = _TransPort(_TransPort.get_socket(self.plugin_path))
        else:
            # At this point we don't know if the user specified an incorrect
            # plug-in in the URI or the daemon isn't started.  We will check
            # the directory for other unix domain sockets.
            if Client._check_daemon_exists():
                raise LsmError(ErrorNumber.PLUGIN_NOT_EXIST,
                               "Plug-in %s not found!" % self.plugin_path)
            else:
                _raise_no_daemon()

        self.__start(uri, plain_text_password, timeout_ms, flags)

    # Synonym for close.
    @_return_requires(None)
    def plugin_unregister(self, flags=FLAG_RSVD):
        """
        Synonym for close.
        """
        self.close(flags)

    # Does an orderly plugin_unregister of the plug-in
    # @param    self    The this pointer
    # @param    flags   Reserved for future use, must be zero.
    @_return_requires(None)
    def close(self, flags=FLAG_RSVD):
        """
        Does an orderly plugin_unregister of the plug-in
        """
        self._tp.rpc('plugin_unregister', _del_self(locals()))
        self._tp.close()
        self._tp = None

    # Retrieves all the available plug-ins
    # @param    field_sep   Field separator
    # @param    flags:      Reserved for future use
    @staticmethod
    @_return_requires([six.string_types[0]])
    def available_plugins(field_sep=':', flags=FLAG_RSVD):
        """
        Retrieves all the available plug-ins

        Return list of strings of available plug-ins with the
        "desc<sep>version"
        """
        rc = []

        if not Client._check_daemon_exists():
            _raise_no_daemon()

        uds_path = Client._plugin_uds_path()

        for root, sub_folders, files in os.walk(uds_path):
            for filename in files:
                uds = os.path.join(root, filename)
                tp = _TransPort(_TransPort.get_socket(uds))
                i, v = tp.rpc('plugin_info', dict(flags=flags))
                rc.append("%s%s%s" % (i, field_sep, v))
                tp.close()

        return rc

    # Sets the timeout for the plug-in
    # @param    self    The this pointer
    # @param    ms      Time-out in ms
    # @param    flags   Reserved for future use, must be zero.
    @_return_requires(None)
    def time_out_set(self, ms, flags=FLAG_RSVD):
        """
        Sets any time-outs for the plug-in (ms)

        Return None on success, else LsmError exception
        """
        return self._tp.rpc('time_out_set', _del_self(locals()))

    # Retrieves the current time-out value.
    # @param    self    The this pointer
    # @param    flags   Reserved for future use, must be zero.
    # @returns  Time-out value
    @_return_requires(int)
    def time_out_get(self, flags=FLAG_RSVD):
        """
        Retrieves the current time-out

        Return time-out in ms, else raise LsmError
        """
        return self._tp.rpc('time_out_get', _del_self(locals()))

    # Retrieves the status of the specified job id.
    # @param    self    The this pointer
    # @param    job_id  The job identifier
    # @param    flags   Reserved for future use, must be zero.
    # @returns A tuple ( status (enumeration), percent_complete,
    # completed item)
    @_return_requires(int, int, _IData)
    def job_status(self, job_id, flags=FLAG_RSVD):
        """
        Returns the stats of the given job.

        Returns a tuple ( status (enumeration), percent_complete,
                            completed item).
        else LsmError exception.
        """
        return self._tp.rpc('job_status', _del_self(locals()))

    # Frees the resources for the specified job id.
    # @param    self    The this pointer
    # @param    job_id  Job id in which to release resource for
    # @param    flags   Reserved for future use, must be zero.
    @_return_requires(None)
    def job_free(self, job_id, flags=FLAG_RSVD):
        """
        Frees resources for a given job number.

        Returns None on success, else raises an LsmError
        """
        return self._tp.rpc('job_free', _del_self(locals()))

    # Gets the capabilities of the array.
    # @param    self    The this pointer
    # @param    system  The system of interest
    # @param    flags   Reserved for future use, must be zero.
    # @returns  Capability object
    @_return_requires(Capabilities)
    def capabilities(self, system, flags=FLAG_RSVD):
        """
        Fetches the capabilities of the array

        Returns a capability object, see data,py for details.
        """
        return self._tp.rpc('capabilities', _del_self(locals()))

    # Gets information about the plug-in
    # @param    self    The this pointer
    # @param    flags   Reserved for future use
    # @returns  Tuple (description, version)
    @_return_requires(six.string_types[0], six.string_types[0])
    def plugin_info(self, flags=FLAG_RSVD):
        """
        Returns a description and version of plug-in
        """
        return self._tp.rpc('plugin_info', _del_self(locals()))

    # Returns an array of pool objects.
    # @param    self            The this pointer
    # @param    search_key      Search key
    # @param    search_value    Search value
    # @param    flags           Reserved for future use, must be zero.
    # @returns An array of pool objects.
    @_return_requires([Pool])
    def pools(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        Returns an array of pool objects.  Pools are used in both block and
        file system interfaces, thus the reason they are in the base class.
        """
        _check_search_key(search_key, Pool.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('pools', _del_self(locals()))

    # Returns an array of system objects.
    # @param    self    The this pointer
    # @param    flags   Reserved for future use, must be zero.
    # @returns An array of system objects.
    @_return_requires([System])
    def systems(self, flags=FLAG_RSVD):
        """
        Returns an array of system objects.  System information is used to
        distinguish resources from on storage array to another when the plug=in
        supports the ability to have more than one array managed by it
        """
        return self._tp.rpc('systems', _del_self(locals()))

    # Changes the read cache percentage for a system.
    # @param    self            The this pointer
    # @param    system          System object to target
    # @param    read_pct        Desired read cache percentage
    # @param    flags           Flags
    # @returns None on success, else raises LsmError
    @_return_requires(None)
    def system_read_cache_pct_update(self, system, read_pct, flags=FLAG_RSVD):
        """
        lsm.Client.system_read_cache_pct_update(self, system, read_pct,
                                                flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            This method updates the read cache percentage on a system if
            supported
        Parameters:
            system(lsm.System)
                an lsm.System object.
            read_pct(int)
                the desired read cache percentage.
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            returns None on success, else raises LsmError on errors.
        SpecialExceptions:
        """
        if sys.version_info[0] == 2:
            if not isinstance(read_pct, (int, long)):
                raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                               "Invalid read_pct, should be an integer")
        else:
            if not isinstance(read_pct, int):
                raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                               "Invalid read_pct, should be an integer")

        if read_pct > 100 or read_pct < 0:
            raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                           "Invalid read_pct, should be in range 0 - 100")
        return self._tp.rpc('system_read_cache_pct_update',
                            _del_self(locals()))

    # Register a user/password for the specified initiator for CHAP
    #  authentication.
    # Note: If you pass an empty user and password the expected behavior is to
    #       remove any authentication for the specified initiator.
    # @param    self            The this pointer
    # @param    init_id         The initiator ID
    # @param    in_user         User for inbound CHAP
    # @param    in_password     Password for inbound CHAP
    # @param    out_user        Outbound username
    # @param    out_password    Outbound password
    # @param    flags   Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(None)
    def iscsi_chap_auth(self, init_id, in_user, in_password,
                        out_user, out_password, flags=FLAG_RSVD):
        """
        Register a user/password for the specified initiator for CHAP
        authentication.
        """
        AccessGroup.initiator_id_verify(init_id,
                                        AccessGroup.INIT_TYPE_ISCSI_IQN,
                                        raise_exception=True)
        return self._tp.rpc('iscsi_chap_auth', _del_self(locals()))

    # Returns an array of volume objects
    # @param    self            The this pointer
    # @param    search_key      Search key to use
    # @param    search_value    Search value
    # @param    flags           Reserved for future use, must be zero.
    # @returns An array of volume objects.
    @_return_requires([Volume])
    def volumes(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        Returns an array of volume objects
        """
        _check_search_key(search_key, Volume.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('volumes', _del_self(locals()))

    # Creates a volume
    # @param    self            The this pointer
    # @param    pool            The pool object to allocate storage from
    # @param    volume_name     The human text name for the volume
    # @param    size_bytes      Size of the volume in bytes
    # @param    provisioning    How the volume is to be provisioned
    # @param    flags           Reserved for future use, must be zero.
    # @returns  A tuple (job_id, new volume), when one is None the other is
    #           valid.
    @_return_requires(six.string_types[0], Volume)
    def volume_create(self, pool, volume_name, size_bytes, provisioning,
                      flags=FLAG_RSVD):
        """
        Creates a volume, given a pool, volume name, size and provisioning

        returns a tuple (job_id, new volume)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('volume_create', _del_self(locals()))

    # Re-sizes a volume
    # @param    self    The this pointer
    # @param    volume  The volume object to re-size
    # @param    new_size_bytes  Size of the volume in bytes
    # @param    flags   Reserved for future use, must be zero.
    # @returns  A tuple (job_id, new re-sized volume), when one is
    #           None the other is valid.
    @_return_requires(six.string_types[0], Volume)
    def volume_resize(self, volume, new_size_bytes, flags=FLAG_RSVD):
        """
        Re-sizes a volume.

        Returns a tuple (job_id, re-sized_volume)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('volume_resize', _del_self(locals()))

    # Replicates a volume from the specified pool.
    # @param    self        The this pointer
    # @param    pool        The pool to re-size from
    # @param    rep_type    Replication type
    #                       (enumeration,see common.data.Volume)
    # @param    volume_src  The volume to replicate
    # @param    name        Human readable name of replicated volume
    # @param    flags       Reserved for future use, must be zero.
    # @returns  A tuple (job_id, new replicated volume), when one is
    #           None the other is valid.
    @_return_requires(six.string_types[0], Volume)
    def volume_replicate(self, pool, rep_type, volume_src, name,
                         flags=FLAG_RSVD):
        """
        Replicates a volume from the specified pool.

        Returns a tuple (job_id, replicated volume)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('volume_replicate', _del_self(locals()))

    # Size of a replicated block.
    # @param    self    The this pointer
    # @param    system  The system to request the rep. block range size from
    # @param    flags   Reserved for future use, must be zero
    # @returns  Size of the replicated block in bytes
    @_return_requires(int)
    def volume_replicate_range_block_size(self, system, flags=FLAG_RSVD):
        """
        Returns the size of a replicated block in bytes.
        """
        return self._tp.rpc('volume_replicate_range_block_size',
                            _del_self(locals()))

    # Replicates a portion of a volume to itself or another volume.
    # @param    self    The this pointer
    # @param    rep_type    Replication type
    #                       (enumeration, see common.data.Volume)
    # @param    volume_src  The volume src to replicate from
    # @param    volume_dest The volume dest to replicate to
    # @param    ranges      An array of Block range objects
    #                       @see lsm.common.data.BlockRange
    # @param    flags       Reserved for future use, must be zero.
    # @returns Job id or None when completed, else raises LsmError on errors.
    @_return_requires(six.string_types[0])
    def volume_replicate_range(self, rep_type, volume_src, volume_dest, ranges,
                               flags=FLAG_RSVD):
        """
        Replicates a portion of a volume to itself or another volume.  The src,
        dest and number of blocks values change with vendor, call
        volume_replicate_range_block_size to get block unit size.

        Returns Job id or None when completed, else raises LsmError on errors.
        """
        return self._tp.rpc('volume_replicate_range', _del_self(locals()))

    # Deletes a volume
    # @param    self    The this pointer
    # @param    volume  The volume object which represents the volume to delete
    # @param    flags   Reserved for future use, must be zero.
    # @returns None on success, else job id.  Raises LsmError on errors.
    @_return_requires(six.string_types[0])
    def volume_delete(self, volume, flags=FLAG_RSVD):
        """
        Deletes a volume.

        Returns None on success, else job id
        """
        return self._tp.rpc('volume_delete', _del_self(locals()))

    # Makes a volume online and available to the host.
    # @param    self    The this pointer
    # @param    volume  The volume to place online
    # @param    flags   Reserved for future use, must be zero.
    # @returns None on success, else raises LsmError
    @_return_requires(None)
    def volume_enable(self, volume, flags=FLAG_RSVD):
        """
        Makes a volume available to the host

        returns None on success, else raises LsmError on errors.
        """
        return self._tp.rpc('volume_enable', _del_self(locals()))

    # Takes a volume offline
    # @param    self    The this pointer
    # @param    volume  The volume object
    # @param    flags   Reserved for future use, must be zero.
    # @returns None on success, else raises LsmError on errors.
    @_return_requires(None)
    def volume_disable(self, volume, flags=FLAG_RSVD):
        """
        Makes a volume unavailable to the host

        returns None on success, else raises LsmError on errors.
        """
        return self._tp.rpc('volume_disable', _del_self(locals()))

    # Returns an array of disk objects
    # @param    self    The this pointer
    # @param    search_key      Search Key
    # @param    search_value    Search value
    # @param    flags   When equal to DISK.FLAG_RETRIEVE_FULL_INFO
    #                   returned objects will contain optional data.
    #                   If not defined, only the mandatory properties will
    #                   be returned.
    # @returns An array of disk objects.
    @_return_requires([Disk])
    def disks(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        Returns an array of disk objects
        """
        _check_search_key(search_key, Disk.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('disks', _del_self(locals()))

    # Access control for allowing an access group to access a volume
    # @param    self            The this pointer
    # @param    access_group    The access group
    # @param    volume          The volume to grant access to
    # @param    flags           Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(None)
    def volume_mask(self, access_group, volume, flags=FLAG_RSVD):
        """
        Allows an access group to access a volume.
        """
        return self._tp.rpc('volume_mask', _del_self(locals()))

    # Revokes access to a volume to initiators in an access group
    # @param    self            The this pointer
    # @param    access_group    The access group
    # @param    volume          The volume to grant access to
    # @param    flags           Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(None)
    def volume_unmask(self, access_group, volume, flags=FLAG_RSVD):
        """
        Revokes access for an access group for a volume
        """
        return self._tp.rpc('volume_unmask', _del_self(locals()))

    # Returns a list of access group objects
    # @param    self    The this pointer
    # @param    search_key      Search Key
    # @param    search_value    Search value
    # @param    flags   Reserved for future use, must be zero.
    # @returns  List of access groups
    @_return_requires([AccessGroup])
    def access_groups(self, search_key=None, search_value=None,
                      flags=FLAG_RSVD):
        """
        Returns a list of access groups
        """
        _check_search_key(search_key, AccessGroup.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('access_groups', _del_self(locals()))

    # Creates an access a group with the specified initiator in it.
    # @param    self                The this pointer
    # @param    name                The initiator group name
    # @param    init_id             Initiator id
    # @param    init_type           Type of initiator (Enumeration)
    # @param    system              Which system to create this group on
    # @param    flags               Reserved for future use, must be zero.
    # @returns AccessGroup on success, else raises LsmError
    @_return_requires(AccessGroup)
    def access_group_create(self, name, init_id, init_type, system,
                            flags=FLAG_RSVD):
        """
        Creates an access group and add the specified initiator id,
        init_type and desired access.
        """
        init_type, init_id = AccessGroup.initiator_id_verify(
            init_id, init_type, raise_exception=True)[1:]
        return self._tp.rpc('access_group_create', _del_self(locals()))

    # Deletes an access group.
    # @param    self            The this pointer
    # @param    access_group    The access group to delete
    # @param    flags           Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(None)
    def access_group_delete(self, access_group, flags=FLAG_RSVD):
        """
        Deletes an access group
        """
        return self._tp.rpc('access_group_delete', _del_self(locals()))

    # Adds an initiator to an access group
    # @param    self            The this pointer
    # @param    access_group    Group to add initiator to
    # @param    init_id         Initiators id
    # @param    init_type       Initiator id type (enumeration)
    # @param    flags           Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(AccessGroup)
    def access_group_initiator_add(self, access_group, init_id, init_type,
                                   flags=FLAG_RSVD):
        """
        Adds an initiator to an access group
        """
        init_type, init_id = AccessGroup.initiator_id_verify(
            init_id, init_type, raise_exception=True)[1:]
        return self._tp.rpc('access_group_initiator_add', _del_self(locals()))

    # Deletes an initiator from an access group
    # @param    self            The this pointer
    # @param    access_group    The access group to remove initiator from
    # @param    init_id         The initiator to remove from the group
    # @param    init_type       Initiator id type (enumeration)
    # @param    flags           Reserved for future use, must be zero.
    # @returns None on success, throws LsmError on errors.
    @_return_requires(AccessGroup)
    def access_group_initiator_delete(self, access_group, init_id, init_type,
                                      flags=FLAG_RSVD):
        """
        Deletes an initiator from an access group
        """
        init_id = AccessGroup.initiator_id_verify(init_id, None,
                                                  raise_exception=True)[2]
        return self._tp.rpc('access_group_initiator_delete',
                            _del_self(locals()))

    # Returns the list of volumes that access group has access to.
    # @param    self            The this pointer
    # @param    access_group    The access group to list volumes for
    # @param    flags           Reserved for future use, must be zero.
    # @returns list of volumes
    @_return_requires([Volume])
    def volumes_accessible_by_access_group(self, access_group,
                                           flags=FLAG_RSVD):
        """
        Returns the list of volumes that access group has access to.
        """
        return self._tp.rpc('volumes_accessible_by_access_group',
                            _del_self(locals()))

    # Returns the list of access groups that have access to the specified
    # volume.
    # @param    self        The this pointer
    # @param    volume      The volume to list access groups for
    # @param    flags       Reserved for future use, must be zero.
    # @returns  list of access groups
    @_return_requires([AccessGroup])
    def access_groups_granted_to_volume(self, volume, flags=FLAG_RSVD):
        """
        Returns the list of access groups that have access to the specified
        volume.
        """
        return self._tp.rpc('access_groups_granted_to_volume',
                            _del_self(locals()))

    # Checks to see if a volume has child dependencies.
    # @param    self    The this pointer
    # @param    volume  The volume to check
    # @param    flags   Reserved for future use, must be zero.
    # @returns True or False
    @_return_requires(bool)
    def volume_child_dependency(self, volume, flags=FLAG_RSVD):
        """
        Returns True if this volume has other volumes which are dependant on
        it. Implies that this volume cannot be deleted or possibly modified
        because it would affect its children.
        """
        return self._tp.rpc('volume_child_dependency', _del_self(locals()))

    # Removes any child dependency.
    # @param    self    The this pointer
    # @param    volume  The volume to remove dependencies for
    # @param    flags   Reserved for future use, must be zero.
    # @returns None if complete, else job id.
    @_return_requires(six.string_types[0])
    def volume_child_dependency_rm(self, volume, flags=FLAG_RSVD):
        """
        If this volume has child dependency, this method call will fully
        replicate the blocks removing the relationship between them.  This
        should return None (success) if volume_child_dependency would return
        False.

        Note:  This operation could take a very long time depending on the size
        of the volume and the number of child dependencies.

        Returns None if complete else job id, raises LsmError on errors.
        """
        return self._tp.rpc('volume_child_dependency_rm', _del_self(locals()))

    # Returns a list of file system objects.
    # @param    self            The this pointer
    # @param    search_key      Search Key
    # @param    search_value    Search value
    # @param    flags           Reserved for future use, must be zero.
    # @returns A list of FS objects.
    @_return_requires([FileSystem])
    def fs(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        Returns a list of file systems on the controller.
        """
        _check_search_key(search_key, FileSystem.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('fs', _del_self(locals()))

    # Deletes a file system
    # @param    self    The this pointer
    # @param    fs      The file system to delete
    # @param    flags   Reserved for future use, must be zero.
    # @returns  None on success, else job id
    @_return_requires(six.string_types[0])
    def fs_delete(self, fs, flags=FLAG_RSVD):
        """
        WARNING: Destructive

        Deletes a file system and everything it contains
        Returns None on success, else job id
        """
        return self._tp.rpc('fs_delete', _del_self(locals()))

    # Re-sizes a file system
    # @param    self            The this pointer
    # @param    fs              The file system to re-size
    # @param    new_size_bytes  The new size of the file system in bytes
    # @param    flags           Reserved for future use, must be zero.
    # @returns tuple (job_id, re-sized file system),
    # When one is None the other is valid
    @_return_requires(six.string_types[0], FileSystem)
    def fs_resize(self, fs, new_size_bytes, flags=FLAG_RSVD):
        """
        Re-size a file system

        Returns a tuple (job_id, re-sized file system)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('fs_resize', _del_self(locals()))

    # Creates a file system.
    # @param    self        The this pointer
    # @param    pool        The pool object to allocate space from
    # @param    name        The human text name for the file system
    # @param    size_bytes  The size of the file system in bytes
    # @param    flags       Reserved for future use, must be zero.
    # @returns  tuple (job_id, file system),
    # When one is None the other is valid
    @_return_requires(six.string_types[0], FileSystem)
    def fs_create(self, pool, name, size_bytes, flags=FLAG_RSVD):
        """
        Creates a file system given a pool, name and size.
        Note: size is limited to 2**64 bytes

        Returns a tuple (job_id, file system)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('fs_create', _del_self(locals()))

    # Clones a file system
    # @param    self            The this pointer
    # @param    src_fs          The source file system to clone
    # @param    dest_fs_name    The destination file system clone name
    # @param    snapshot        Optional, create clone from previous snapshot
    # @param    flags           Reserved for future use, must be zero.
    # @returns tuple (job_id, file system)
    @_return_requires(six.string_types[0], FileSystem)
    def fs_clone(self, src_fs, dest_fs_name, snapshot=None, flags=FLAG_RSVD):
        """
        Creates a thin, point in time read/writable copy of src to dest.
        Optionally uses snapshot as backing of src_fs

        Returns a tuple (job_id, file system)
        Note: Tuple return values are mutually exclusive, when one
        is None the other must be valid.
        """
        return self._tp.rpc('fs_clone', _del_self(locals()))

    # Clones an individual file or files on the specified file system
    # @param    self            The this pointer
    # @param    fs              The file system the files are on
    # @param    src_file_name   The source file name
    # @param    dest_file_name  The dest. file name
    # @param    snapshot        Optional, the snapshot to base clone source
    #                                       file from
    # @param    flags           Reserved for future use, must be zero.
    # @returns  None on success, else job id
    @_return_requires(six.string_types[0])
    def fs_file_clone(self, fs, src_file_name, dest_file_name, snapshot=None,
                      flags=FLAG_RSVD):
        """
        Creates a thinly provisioned clone of src to dest.
        Note: Source and Destination are required to be on same filesystem and
        all directories in destination path need to exist.

        Returns None on success, else job id
        """
        return self._tp.rpc('fs_file_clone', _del_self(locals()))

    # Returns a list of snapshots
    # @param    self    The this pointer
    # @param    fs      The file system
    # @param    flags   Reserved for future use, must be zero.
    # @returns  a list of snapshot objects.
    @_return_requires([FsSnapshot])
    def fs_snapshots(self, fs, flags=FLAG_RSVD):
        """
        Returns a list of snapshot names for the supplied file system
        """
        return self._tp.rpc('fs_snapshots', _del_self(locals()))

    # Creates a snapshot (Point in time read only copy)
    # @param    self            The this pointer
    # @param    fs              The file system to snapshot
    # @param    snapshot_name   The human readable snapshot name
    # @param    flags           Reserved for future use, must be zero.
    # @returns tuple (job_id, snapshot)
    @_return_requires(six.string_types[0], FsSnapshot)
    def fs_snapshot_create(self, fs, snapshot_name, flags=FLAG_RSVD):
        """
        Snapshot is a point in time read-only copy

        Create a snapshot on the chosen file system.

        Returns a tuple (job_id, snapshot)
        Notes:
        - Snapshot name may not match what was passed in
          (depends on array implementation)
        - Tuple return values are mutually exclusive, when one
          is None the other must be valid.
        """
        return self._tp.rpc('fs_snapshot_create', _del_self(locals()))

    # Deletes a snapshot
    # @param    self        The this pointer
    # @param    fs          The filesystem the snapshot it for
    # @param    snapshot    The specific snap shot to delete
    # @param    flags       Reserved for future use, must be zero.
    # @returns  None on success, else job id
    @_return_requires(six.string_types[0])
    def fs_snapshot_delete(self, fs, snapshot, flags=FLAG_RSVD):
        """
        Frees the re-sources for the given snapshot on the supplied filesystem.

        Returns None on success else job id, LsmError exception on error
        """
        return self._tp.rpc('fs_snapshot_delete', _del_self(locals()))

    # Reverts a snapshot
    # @param    self            The this pointer
    # @param    fs              The file system object to restore snapshot for
    # @param    snapshot        The snapshot file to restore back too
    # @param    files           The specific files to restore
    # @param    restore_files   Individual files to restore
    # @param    all_files       Set to True if all files should be restored
    #                           back
    # @param    flags           Reserved for future use, must be zero.
    # @return None on success, else job id
    @_return_requires(six.string_types[0])
    def fs_snapshot_restore(self, fs, snapshot, files, restore_files,
                            all_files=False, flags=FLAG_RSVD):
        """
        WARNING: Destructive!

        Reverts a file-system or just the specified files from the snapshot.
        If a list of files is supplied but the array cannot restore just them
        then the operation will fail with an LsmError raised.  If files == None
        and all_files = True then all files on the file-system are restored.

        Restore_file if None none must be the same length as files with each
        index in each list referring to the associated file.

        Returns None on success, else job id, LsmError exception on error
        """
        return self._tp.rpc('fs_snapshot_restore', _del_self(locals()))

    # Checks to see if a file system has child dependencies.
    # @param    fs      The file system to check
    # @param    files   The files to check (optional)
    # @param    flags   Reserved for future use, must be zero.
    # @returns True or False
    @_return_requires(bool)
    def fs_child_dependency(self, fs, files, flags=FLAG_RSVD):
        """
        Returns True if the specified filesystem or specified file on this
        file system has child dependencies.  This implies that this filesystem
        or specified file on this file system cannot be deleted or possibly
        modified because it would affect its children.
        """
        return self._tp.rpc('fs_child_dependency', _del_self(locals()))

    # Removes child dependencies from a FS or specific file.
    # @param    self    The this pointer
    # @param    fs      The file system to remove child dependencies for
    # @param    files   The list of files to remove child dependencies (opt.)
    # @param    flags   Reserved for future use, must be zero.
    # @returns None if complete, else job id.
    @_return_requires(six.string_types[0])
    def fs_child_dependency_rm(self, fs, files, flags=FLAG_RSVD):
        """
        If this filesystem or specified file on this filesystem has child
        dependency this method will fully replicate the blocks removing the
        relationship between them.  This should return None(success) if
        fs_child_dependency would return False.

        Note:  This operation could take a very long time depending on the size
        of the filesystem and the number of child dependencies.

        Returns None if completed, else job id.  Raises LsmError on errors.
        """
        return self._tp.rpc('fs_child_dependency_rm', _del_self(locals()))

    # Returns a list of all the NFS client authentication types.
    # @param    self    The this pointer
    # @param    flags   Reserved for future use, must be zero.
    # @returns  An array of client authentication types.
    @_return_requires([six.string_types[0]])
    def export_auth(self, flags=FLAG_RSVD):
        """
        What types of NFS client authentication are supported.
        """
        return self._tp.rpc('export_auth', _del_self(locals()))

    # Returns a list of all the exported file systems
    # @param    self    The this pointer
    # @param    search_key      Search Key
    # @param    search_value    Search value
    # @param    flags   Reserved for future use, must be zero.
    # @returns An array of export objects
    @_return_requires([NfsExport])
    def exports(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        Get a list of all exported file systems on the controller.
        """
        _check_search_key(search_key, NfsExport.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('exports', _del_self(locals()))

    # Exports a FS as specified in the export.
    # @param    self            The this pointer
    # @param    fs_id           The FS ID to export
    # @param    export_path     The export path (Set to None for array to pick)
    # @param    root_list       List of hosts with root access
    # @param    rw_list         List of hosts with read/write access
    # @param    ro_list         List of hosts with read only access
    # @param    anon_uid        UID to map to anonymous
    # @param    anon_gid        GID to map to anonymous
    # @param    auth_type       NFS client authentication type
    # @param    options         Options to pass to plug-in
    # @param    flags           Reserved for future use, must be zero.
    # @returns NfsExport on success, else raises LsmError
    @_return_requires(NfsExport)
    def export_fs(self, fs_id, export_path, root_list, rw_list, ro_list,
                  anon_uid=NfsExport.ANON_UID_GID_NA,
                  anon_gid=NfsExport.ANON_UID_GID_NA,
                  auth_type=None, options=None, flags=FLAG_RSVD):
        """
        Exports a filesystem as specified in the arguments
        """
        if set(rw_list) - (set(rw_list) - set(ro_list)):
            raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                           "Host cannot both in rw_list and ro_list.")
        return self._tp.rpc('export_fs', _del_self(locals()))

    # Removes the specified export
    # @param    self    The this pointer
    # @param    export  The export to remove
    # @param    flags   Reserved for future use, must be zero.
    # @returns None on success, else raises LsmError
    @_return_requires(None)
    def export_remove(self, export, flags=FLAG_RSVD):
        """
        Removes the specified export
        """
        return self._tp.rpc('export_remove', _del_self(locals()))

    # Returns a list of target ports
    # @param    self    The this pointer
    # @param    search_key      The key to search against
    # @param    search_value    The value to search for
    # @param    flags           Reserved for future use, must be zero
    # @returns List of target ports, else raises LsmError
    @_return_requires([TargetPort])
    def target_ports(self, search_key=None, search_value=None,
                     flags=FLAG_RSVD):
        """
        Returns a list of target ports
        """
        _check_search_key(search_key, TargetPort.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('target_ports', _del_self(locals()))

    # Returns the RAID information of certain volume
    # @param    self    The this pointer
    # @param    volume  The volume to retrieve RAID information for
    # @param    flags   Reserved for future use, must be zero
    # @returns [raid_type, strip_size, disk_count, min_io_size, opt_io_size],
    #           else raises LsmError
    @_return_requires([int, int, int, int, int])
    def volume_raid_info(self, volume, flags=FLAG_RSVD):
        """Query the RAID information of certain volume.

        New in version 1.2.

        Query the RAID type, strip size, extents count, minimum I/O size,
        optimal I/O size of given volume.

        This method requires this capability:
            lsm.Capabilities.VOLUME_RAID_INFO

        Args:
            volume (Volume object): Volume to query
            flags (int): Reserved for future use. Should be set as
                lsm.Client.FLAG_RSVD
        Returns:
            [raid_type, strip_size, disk_count, min_io_size, opt_io_size]

            raid_type (int): RAID Type of requested volume.
                Could be one of these values:
                    Volume.RAID_TYPE_RAID0
                        Stripe
                    Volume.RAID_TYPE_RAID1
                        Two disks Mirror
                    Volume.RAID_TYPE_RAID3
                        Byte-level striping with dedicated parity
                    Volume.RAID_TYPE_RAID4
                        Block-level striping with dedicated parity
                    Volume.RAID_TYPE_RAID5
                        Block-level striping with distributed parity
                    Volume.RAID_TYPE_RAID6
                        Block-level striping with two distributed parities,
                        aka, RAID-DP
                    Volume.RAID_TYPE_RAID10
                        Stripe of mirrors
                    Volume.RAID_TYPE_RAID15
                        Parity of mirrors
                    Volume.RAID_TYPE_RAID16
                        Dual parity of mirrors
                    Volume.RAID_TYPE_RAID50
                        Stripe of parities
                    Volume.RAID_TYPE_RAID60
                        Stripe of dual parities
                    Volume.RAID_TYPE_RAID51
                        Mirror of parities
                    Volume.RAID_TYPE_RAID61
                        Mirror of dual parities
                    Volume.RAID_TYPE_JBOD
                        Just bunch of disks, no parity, no striping.
                    Volume.RAID_TYPE_UNKNOWN
                        The plugin failed to detect the volume's RAID type.
                    Volume.RAID_TYPE_MIXED
                        This volume contains multiple RAID settings.
                    Volume.RAID_TYPE_OTHER
                        Vendor specific RAID type
            strip_size(int): The size of strip on each disk or other storage
                extent.
                For RAID1/JBOD, it should be set as sector size.
                If plugin failed to detect strip size, it should be set
                as Volume.STRIP_SIZE_UNKNOWN(0).
            disk_count(int): The count of disks used for assembling the RAID
                group(s) where this volume allocated from.
                For any RAID system using the slice of disk, this value
                indicate how many disk slices are used for the RAID.
                For example, on LVM RAID, the 'disk_count' here indicate the
                count of PVs used for certain volume.
                Another example, on EMC VMAX, the 'disk_count' here indicate
                how many hyper volumes are used for this volume.
                For any RAID system using remote LUN for data storing, each
                remote LUN should be count as a disk.
                If the plugin failed to detect disk_count, it should be set
                as Volume.DISK_COUNT_UNKNOWN(0).
            min_io_size(int): The minimum I/O size, device preferred I/O
                size for random I/O. Any I/O size not equal to a multiple
                of this value may get significant speed penalty.
                Normally it refers to strip size of each disk(extent).
                If plugin failed to detect min_io_size, it should try these
                values in the sequence of:
                logical sector size -> physical sector size ->
                Volume.MIN_IO_SIZE_UNKNOWN(0).
            opt_io_size(int): The optimal I/O size, device preferred I/O
                size for sequential I/O. Normally it refers to RAID group
                stripe size.
                If plugin failed to detect opt_io_size, it should be set
                to Volume.OPT_IO_SIZE_UNKNOWN(0).
        Raises:
            LsmError:
                ErrorNumber.NO_SUPPORT
                    No support.
        """
        return self._tp.rpc('volume_raid_info', _del_self(locals()))

    # Query the membership information of specified pool
    # @param    self    The this pointer
    # @param    pool    The pool to query
    # @param    flags   Reserved for future use, must be zero
    # @returns  [raid_type, member_type, [member_ids]], lsmError on error
    @_return_requires([int, int, [six.string_types[0]]])
    def pool_member_info(self, pool, flags=FLAG_RSVD):
        """
        lsm.Client.pool_member_info(self, pool, flags=lsm.Client.FLAG_RSVD)

        Version:
            1.2
        Usage:
            Query the membership information of certain pool:
                RAID type, member type and member ids.
            Currently, LibStorageMgmt supports two types of pool:
                * Sub-pool -- Pool.MEMBER_TYPE_POOL
                    Pool space is allocated from parent pool.
                    Example:
                        * NetApp ONTAP volume

                * Disk RAID pool -- Pool.MEMBER_TYPE_DISK
                    Pool is a RAID group assembled by disks.
                    Example:
                        * LSI MegaRAID disk group
                        * EMC VNX pool
                        * NetApp ONTAP aggregate
        Parameters:
            pool (lsm.Pool object)
                Pool to query
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            [raid_type, member_type, member_ids]

            raid_type (int)
                RAID Type of requested pool.
                    Could be one of these values:
                        Volume.RAID_TYPE_RAID0
                            Stripe
                        Volume.RAID_TYPE_RAID1
                            Two disks Mirror
                        Volume.RAID_TYPE_RAID3
                            Byte-level striping with dedicated parity
                        Volume.RAID_TYPE_RAID4
                            Block-level striping with dedicated parity
                        Volume.RAID_TYPE_RAID5
                            Block-level striping with distributed parity
                        Volume.RAID_TYPE_RAID6
                            Block-level striping with two distributed parities,
                            aka, RAID-DP
                        Volume.RAID_TYPE_RAID10
                            Stripe of mirrors
                        Volume.RAID_TYPE_RAID15
                            Parity of mirrors
                        Volume.RAID_TYPE_RAID16
                            Dual parity of mirrors
                        Volume.RAID_TYPE_RAID50
                            Stripe of parities Volume.RAID_TYPE_RAID60
                            Stripe of dual parities
                        Volume.RAID_TYPE_RAID51
                            Mirror of parities
                        Volume.RAID_TYPE_RAID61
                            Mirror of dual parities
                        Volume.RAID_TYPE_JBOD
                            Just bunch of disks, no parity, no striping.
                        Volume.RAID_TYPE_UNKNOWN
                            The plugin failed to detect the volume's RAID type.
                        Volume.RAID_TYPE_MIXED
                            This pool contains multiple RAID settings.
                        Volume.RAID_TYPE_OTHER
                            Vendor specific RAID type
            member_type (int)
                Could be one of these values:
                    Pool.MEMBER_TYPE_POOL
                        Current pool(also known as sub-pool) is allocated from
                        other pool(parent pool). The 'raid_type' will set to
                        RAID_TYPE_OTHER unless certain RAID system support RAID
                        using space of parent pools.
                    Pool.MEMBER_TYPE_DISK
                        Pool is created from RAID group using whole disks.
                    Pool.MEMBER_TYPE_OTHER
                        Vendor specific RAID member type.
                    Pool.MEMBER_TYPE_UNKNOWN
                        Plugin failed to detect the RAID member type.
            member_ids (list of strings)
                When 'member_type' is Pool.MEMBER_TYPE_POOL,
                the 'member_ids' will contain a list of parent Pool IDs.
                When 'member_type' is Pool.MEMBER_TYPE_DISK,
                the 'member_ids' will contain a list of disk IDs.
                When 'member_type' is Pool.MEMBER_TYPE_OTHER or
                Pool.MEMBER_TYPE_UNKNOWN, the member_ids should be an empty
                list.
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
                ErrorNumber.NOT_FOUND_POOL
        Capability:
            lsm.Capabilities.POOL_MEMBER_INFO
        """
        return self._tp.rpc('pool_member_info', _del_self(locals()))

    # Queries all the supported RAID types and stripe sizes which could be
    #  used for input into volume_raid_create
    # @param    self    The this pointer
    # @param    system  System (raid) card to query
    # @param    flags   Reserved for future use, must be zero
    # @returns  [raid_types, strip_sizes], lsmError on error
    @_return_requires([[int], [int]])
    def volume_raid_create_cap_get(self, system, flags=FLAG_RSVD):
        """
        lsm.Client.volume_raid_create_cap_get(
            self, system, flags=lsm.Client.FLAG_RSVD)

        Version:
            1.2
        Usage:
            This method is dedicated to local hardware RAID cards.
            Query out all supported RAID types and strip sizes which could
            be used by lsm.Client.volume_raid_create() method.
        Parameters:
            system (lsm.System)
                Instance of lsm.System
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            [raid_types, strip_sizes]
                raid_types ([int])
                    List of integer, possible values are:
                        Volume.RAID_TYPE_RAID0
                        Volume.RAID_TYPE_RAID1
                        Volume.RAID_TYPE_RAID5
                        Volume.RAID_TYPE_RAID6
                        Volume.RAID_TYPE_RAID10
                        Volume.RAID_TYPE_RAID50
                        Volume.RAID_TYPE_RAID60
                strip_sizes ([int])
                    List of integer. Stripe size in bytes.
        SpecialExceptions:
            LsmError
                lsm.ErrorNumber.NO_SUPPORT
                    Method not supported.
        Sample:
            lsm_client = lsm.Client('sim://')
            lsm_sys = lsm_client.systems()[0]
            disks = lsm_client.disks(
                search_key='system_id', search_value=lsm_sys.id)

            free_disks = [d for d in disks if d.status == Disk.STATUS_FREE]
            supported_raid_types, supported_strip_sizes = \
                lsm_client.volume_raid_create_cap_get(lsm_sys)
            new_vol = lsm_client.volume_raid_create(
                'test_volume_raid_create', supported_raid_types[0],
                free_disks, supported_strip_sizes[0])

        Capability:
            lsm.Capabilities.VOLUME_CREATE_RAID
                This method is mandatory when volume_raid_create() is
                supported.
        """
        return self._tp.rpc('volume_raid_create_cap_get', _del_self(locals()))

    # Create a disk RAID pool and allocate entire storage space to new volume
    # @param    self            The this pointer
    # @param    name            Name of the raid to be created
    # @param    raid_type       The raid type (enumerated) of the raid group
    # @param    disks           List of disk objects to use
    # @param    strip_size      Strip size in bytes
    # @param    flags           Flags
    # @returns  the newly created volume, lsmError on errors
    @_return_requires(Volume)
    def volume_raid_create(self, name, raid_type, disks, strip_size,
                           flags=FLAG_RSVD):
        """
        lsm.Client.volume_raid_create(self, name, raid_type, disks,
                                      strip_size, flags=lsm.Client.FLAG_RSVD)

        Version:
            1.2
        Usage:
            This method is dedicated to local hardware RAID cards.
            Create a disk RAID pool and allocate entire storage space to
            new volume using requested volume name.
            When dealing with RAID10, 50 or 60, the first half part of
            'disks' will be located in one bottom layer RAID group.
            The new volume and new pool will created within the same system
            of provided disks.
            This method does not allow duplicate call, when duplicate call
            was issued, LsmError with ErrorNumber.DISK_NOT_FREE will be raise.
            User should check disk.status for Disk.STATUS_FREE before
            invoking this method.
        Parameters:
            name (string)
                The name for new volume.
                The requested volume name might be ignored due to restriction
                of hardware RAID vendors.
                The pool name will be automatically chosen by plugin.
            raid_type (int)
                The RAID type for the RAID group, possible values are:
                    Volume.RAID_TYPE_RAID0
                    Volume.RAID_TYPE_RAID1
                    Volume.RAID_TYPE_RAID5
                    Volume.RAID_TYPE_RAID6
                    Volume.RAID_TYPE_RAID10
                    Volume.RAID_TYPE_RAID15
                    Volume.RAID_TYPE_RAID16
                    Volume.RAID_TYPE_RAID50
                Please check volume_raid_create_cap_get() returns to get
                supported all raid types of current hardware RAID card.
            disks ([lsm.Disks,])
                A list of lsm.Disk objects. Free disks used for new RAID group.
            strip_size (int)
                The size in bytes of strip.
                When setting strip_size to Volume.VCR_STRIP_SIZE_DEFAULT, it
                allow hardware RAID cards to choose their default value.
                Please use volume_raid_create_cap_get() method to get all
                supported strip size of current hardware RAID card.
                The Volume.VCR_STRIP_SIZE_DEFAULT is always supported when
                lsm.Capabilities.VOLUME_CREATE_RAID is supported.
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            lsm.Volume
                The lsm.Volume object for newly created volume.
        SpecialExceptions:
            LsmError
                lsm.ErrorNumber.NO_SUPPORT
                    Method not supported or RAID type not supported.
                lsm.ErrorNumber.DISK_NOT_FREE
                    Disk is not in Disk.STATUS_FREE status.
                lsm.ErrorNumber.NOT_FOUND_DISK
                    Disk not found
                lsm.ErrorNumber.INVALID_ARGUMENT
                    1. Invalid input argument data.
                    2. Disks are not from the same system.
                    3. Disks are not from the same enclosure.
                    4. Invalid strip_size.
                    5. Disk count are meet the minimum requirement:
                        RAID1: len(disks) == 2
                        RAID5: len(disks) >= 3
                        RAID6: len(disks) >= 4
                        RAID10: len(disks) % 2 == 0 and len(disks) >= 4
                        RAID50: len(disks) % 2 == 0 and len(disks) >= 6
                        RAID60: len(disks) % 2 == 0 and len(disks) >= 8
                lsm.ErrorNumber.NAME_CONFLICT
                    Requested name is already be used by other volume.
        Sample:
            lsm_client = lsm.Client('sim://')
            disks = lsm_client.disks()
            free_disks = [d for d in disks if d.status == Disk.STATUS_FREE]
            new_vol = lsm_client.volume_raid_create(
                'raid0_vol1', Volume.RAID_TYPE_RAID0, free_disks)
        Capability:
            lsm.Capabilities.VOLUME_CREATE_RAID
                Indicate current system support volume_raid_create() method.
                At least one RAID type should be supported.
                The strip_size == Volume.VCR_STRIP_SIZE_DEFAULT is supported.
        """
        if len(disks) == 0:
            raise LsmError(
                ErrorNumber.INVALID_ARGUMENT,
                "Illegal input disks argument: no disk included")

        if raid_type == Volume.RAID_TYPE_RAID1 and len(disks) != 2:
            raise LsmError(
                ErrorNumber.INVALID_ARGUMENT,
                "Illegal input disks argument: RAID 1 only allow 2 disks")

        if raid_type == Volume.RAID_TYPE_RAID5 and len(disks) < 3:
            raise LsmError(
                ErrorNumber.INVALID_ARGUMENT,
                "Illegal input disks argument: RAID 5 require 3 or more disks")

        if raid_type == Volume.RAID_TYPE_RAID6 and len(disks) < 4:
            raise LsmError(
                ErrorNumber.INVALID_ARGUMENT,
                "Illegal input disks argument: RAID 6 require 4 or more disks")

        if raid_type == Volume.RAID_TYPE_RAID10:
            if len(disks) % 2 or len(disks) < 4:
                raise LsmError(
                    ErrorNumber.INVALID_ARGUMENT,
                    "Illegal input disks argument: "
                    "RAID 10 require even disks count and 4 or more disks")

        if raid_type == Volume.RAID_TYPE_RAID50:
            if len(disks) % 2 or len(disks) < 6:
                raise LsmError(
                    ErrorNumber.INVALID_ARGUMENT,
                    "Illegal input disks argument: "
                    "RAID 50 require even disks count and 6 or more disks")

        if raid_type == Volume.RAID_TYPE_RAID60:
            if len(disks) % 2 or len(disks) < 8:
                raise LsmError(
                    ErrorNumber.INVALID_ARGUMENT,
                    "Illegal input disks argument: "
                    "RAID 60 require even disks count and 8 or more disks")

        return self._tp.rpc('volume_raid_create', _del_self(locals()))

    # Enable the IDENT LED for a volume.
    # @param    self            The this pointer
    # @param    volume          Volume object to target
    # @param    flags           Flags
    # @returns None on success, else raises LsmError
    @_return_requires(None)
    def volume_ident_led_on(self, volume, flags=FLAG_RSVD):
        """
        lsm.Client.volume_ident_led_on(self, volume,
                                       flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            This method enables the IDENT LED on a volume if supported
        Parameters:
            volume (lsm.Volume)
                An lsm.Volume object.
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            returns None on success, else raises LsmError on errors.
        SpecialExceptions:
        """
        return self._tp.rpc('volume_ident_led_on', _del_self(locals()))

    # Disable the IDENT LED for a volume.
    # @param    self            The this pointer
    # @param    volume          Volume object to target
    # @param    flags           Flags
    # @returns None on success, else raises LsmError
    @_return_requires(None)
    def volume_ident_led_off(self, volume, flags=FLAG_RSVD):
        """
        lsm.Client.volume_ident_led_off(self, volume,
                                          flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            This method disables the IDENT LED on a volume if supported
        Parameters:
            volume (lsm.Volume)
                An lsm.Volume object.
            flags (int)
                Optional. Reserved for future use.
                Should be set as lsm.Client.FLAG_RSVD.
        Returns:
            returns None on success, else raises LsmError on errors.
        SpecialExceptions:
        """
        return self._tp.rpc('volume_ident_led_off', _del_self(locals()))

    @_return_requires([Battery])
    def batteries(self, search_key=None, search_value=None, flags=FLAG_RSVD):
        """
        lsm.Client.batteries(self, search_key=None, search_value=None,
                             flags=lsm.Client.FLAG_RSVD)
        Version:
            1.3
        Usage:
            Query batteries. When present, super capacitors will also be
            included.
        Parameters:
            search_key (string, optional)
                The key name for the search. Valid search keys are stored in
                lsm.Battery.SUPPORTED_SEARCH_KEYS
            search_value (string, optional)
                The value of search_key to match.
            flags (int, optional):
                Reserved for future use. Should be set as lsm.Client.FLAG_RSVD
        Returns:
            [lsm.Battery]

            lsm.Battery (object)
                lsm.Battery.id (string)
                    Unique ID for this cache hardware.
                lsm.Battery.name (string)
                    Human friendly name, might include physical location, model
                    name and etc.
                lsm.Battery.type (int)
                    The hardware type of cache. Could be one of these values:
                        lsm.Battery.TYPE_CHEMICAL
                            Chemical battery.
                        lsm.Battery.TYPE_CAPACITOR
                            Supper capacitor.
                        lsm.Battery.TYPE_OTHER
                            Vendor specific battery type.
                        lsm.Battery.TYPE_UNKNOWN
                            Unknown type.
                lsm.Battery.status (int, bitmap)
                    Could be any combination of these values:
                        lsm.Battery.STATUS_OK
                            Battery is fully charged, health and not in use
                            currently.
                        lsm.Battery.STATUS_DISCHARGING
                            Battery is in use.
                        lsm.Battery.STATUS_CHARGING
                            Battery is charging.
                        lsm.Battery.STATUS_LEARNING
                            Battery is calibrating itself by discharging
                            battery and recharging again.
                        lsm.Battery.STATUS_DEGRADED
                            Battery is in degraded mode, need attention.
                            For example, battery is near end of life.
                        lsm.Battery.STATUS_ERROR
                            Battery is having hardware error or end of life.
                        lsm.Battery.STATUS_OTHER
                            Vendor specific status.
                        lsm.Battery.STATUS_UNKNOWN
                            Unknown.
                lsm.Battery.system_id (string)
                    The id of system which current battery belong to.
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
        Capability:
            lsm.Capabilities.BATTERIES
        """
        _check_search_key(search_key, Battery.SUPPORTED_SEARCH_KEYS)
        return self._tp.rpc('batteries', _del_self(locals()))

    @_return_requires([int, int, int, int, int])
    def volume_cache_info(self, volume, flags=FLAG_RSVD):
        """
        lsm.Client.volume_cache_info(self, volume,
                                     flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            Query RAM cache setting and status of specified volume on read
            and write I/O.
        Parameters:
            volume (Lsm.Volume)
                The lsm.Volume instance.
            flags (int, optional):
                Reserved for future use. Should be set as lsm.Client.FLAG_RSVD
        Returns:
            [int, int, int, int, int]
            write_cache_policy (int)
                The write cache policy. Valid values are:
                    * lsm.Volume.WRITE_CACHE_POLICY_WRITE_BACK
                        The storage system will use write back mode if cache
                        hardware found.
                    * lsm.Volume.WRITE_CACHE_POLICY_AUTO
                        The controller will use write back mode when
                        battery/capacitor is in good health, otherwise,
                        write through mode.
                    * lsm.Volume.WRITE_CACHE_POLICY_WRITE_THROUGH
                        The storage system will use write through mode.
                    * lsm.Volume.WRITE_CACHE_POLICY_UNKNOWN
                        Plugin failed to detect this setting.

            write_cache_status (int)
                The status of write cache. Valid values are:
                    * lsm.Volume.WRITE_CACHE_STATUS_WRITE_THROUGH
                    * lsm.Volume.WRITE_CACHE_STATUS_WRITE_BACK
                    * lsm.Volume.WRITE_CACHE_STATUS_UNKNOWN

            read_cache_policy (int)
                The policy for read cache. Valid values are:
                    * lsm.Volume.READ_CACHE_POLICY_ENABLED
                        Read cache is enabled, when reading I/O on previous
                        unchanged written I/O or read I/O in cache will be
                        returned to I/O initiator immediately without checking
                        backing store(normally disk).
                    * lsm.Volume.READ_CACHE_POLICY_DISABLED
                        Read cache is disabled.
                    * lsm.Volume.READ_CACHE_POLICY_UNKNOWN
                        Plugin failed to detect the read cache policy.
            read_cache_status (int)
                The status of read cache. Valid values are:
                    * lsm.Volume.READ_CACHE_STATUS_ENABLED
                    * lsm.Volume.READ_CACHE_STATUS_DISABLED
                    * lsm.Volume.READ_CACHE_STATUS_UNKNOWN
            physical_disk_cache (int)
                Whether physical disk's cache is enabled or not.
                Please be advised, HDD's physical disk ram cache might be not
                protected by storage system's battery or capacitor on sudden
                power loss, you could lose data if a power failure occurs
                during a write process.
                For SSD's physical disk cache, please check with the vendor of
                your hardware RAID card and SSD disk.
                Valid values are:
                    * lsm.Volume.PHYSICAL_DISK_CACHE_ENABLED
                        Physical disk cache enabled.
                    * lsm.Volume.PHYSICAL_DISK_CACHE_DISABLED
                        Physical disk cache disabled.
                    * lsm.Volume.PHYSICAL_DISK_CACHE_USE_DISK_SETTING
                        Physical disk cache is determined by the disk vendor
                        via physical disks' SCSI caching mode page(0x08 page).
                        It is strongly suggested to change this value to
                        lsm.Volume.PHYSICAL_DISK_CACHE_ENABLED or
                        lsm.Volume.PHYSICAL_DISK_CACHE_DISABLED
                    * lsm.Volume.PHYSICAL_DISK_CACHE_UNKNOWN
                        Plugin failed to detect the physical disk status.
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
                ErrorNumber.NOT_FOUND_VOLUME
        Capability:
            lsm.Capabilities.VOLUME_CACHE_INFO
        """
        return self._tp.rpc('volume_cache_info', _del_self(locals()))

    @_return_requires(None)
    def volume_physical_disk_cache_update(self, volume, pdc, flags=FLAG_RSVD):
        """
        lsm.Client.volume_physical_disk_cache_update(self, volume, pdc,
                                                     flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            Change the setting of RAM physical disk cache of specified volume.
            On some product(like HPE SmartArray), this action will be effective
            at system level which means that even you are requesting a change
            on a specified volume, this change will apply to all other volumes
            on the same controller(system).
        Parameters:
            volume (Lsm.Volume)
                The lsm.Volume instance.
            pdc (int)
                lsm.Volume.PHYSICAL_DISK_CACHE_ENABLED
                    Enable physical disk cache.
                lsm.Volume.PHYSICAL_DISK_CACHE_DISABLED
                    Disable physical disk cache
            flags (int, optional):
                Reserved for future use. Should be set as lsm.Client.FLAG_RSVD
        Returns:
            N/A
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
                    You might also get NO_SUPPORT error when trying
                    to change SSD physical disk cache on MegaRAID.
                ErrorNumber.NOT_FOUND_VOLUME
        Capability:
            lsm.Capabilities.VOLUME_PHYSICAL_DISK_CACHE_SET
                Allow changing physical disk cache.
            lsm.Capabilities.VOLUME_PHYSICAL_DISK_CACHE_SET_SYSTEM_LEVEL
                Indicate that this action will change system settings which
                are effective on all volumes in this storage system.
                For example, on HPE SmartArray, the physical disk cache
                setting is a controller level setting.
        """
        if (pdc != Volume.PHYSICAL_DISK_CACHE_ENABLED) and \
           (pdc != Volume.PHYSICAL_DISK_CACHE_DISABLED):
            raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                           "Argument pdc should be "
                           "Volume.PHYSICAL_DISK_CACHE_ENABLED or "
                           "Volume.PHYSICAL_DISK_CACHE_DISABLED")

        return self._tp.rpc('volume_physical_disk_cache_update',
                            _del_self(locals()))

    @_return_requires(None)
    def volume_write_cache_policy_update(self, volume, wcp, flags=FLAG_RSVD):
        """
        lsm.Client.volume_write_cache_policy_update(self, volume, wcp,
                                                    flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            Change the RAM write cache policy on specified volume.
            If lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_IMPACT_READ
            is supported(e.g. HPE SmartArray), the changes on write cache
            policy might also impact read cache policy.
            If lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_WB_IMPACT_OTHER
            is supported(e.g. HPE SmartArray), changing write cache policy
            to write back mode might impact other volumes in the same system.
        Parameters:
            volume (Lsm.Volume)
                The lsm.Volume instance.
            wcp (int)
                Could be one of these value:
                    * lsm.Volume.WRITE_CACHE_POLICY_WRITE_BACK
                        Change to write back mode.
                    * lsm.Volume.WRITE_CACHE_POLICY_AUTO
                        Change to auto mode: use write back mode when
                        battery/capacitor is healthy, otherwise use write
                        through.
                    * lsm.Volume.WRITE_CACHE_POLICY_WRITE_THROUGH
                        Change to write through mode.
            flags (int, optional):
                Reserved for future use. Should be set as lsm.Client.FLAG_RSVD
        Returns:
            N/A
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
                ErrorNumber.NOT_FOUND_VOLUME
        Capability:
            lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_WRITE_BACK
                Allow changing to always mode.
            lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_AUTO
                Allow changing to auto mode.
            lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_WRITE_THROUGH
                Allow changing to disable mode.
            lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_IMPACT_READ
                Indicate this action might impact read cache policy.
            lsm.Capabilities.VOLUME_WRITE_CACHE_POLICY_SET_WB_IMPACT_OTHER
                Indicate that changing to write back mode might impact other
                volumes. For example, on HPE SmartArray, changing to write back
                mode will change all other volumes with auto write cache policy
                to write back mode.
        """
        if wcp != Volume.WRITE_CACHE_POLICY_WRITE_BACK and \
           wcp != Volume.WRITE_CACHE_POLICY_AUTO and \
           wcp != Volume.WRITE_CACHE_POLICY_WRITE_THROUGH:
            raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                           "Argument wcp should be "
                           "Volume.WRITE_CACHE_POLICY_WRITE_BACK or "
                           "Volume.WRITE_CACHE_POLICY_AUTO or "
                           "Volume.WRITE_CACHE_POLICY_WRITE_THROUGH")
        return self._tp.rpc('volume_write_cache_policy_update',
                            _del_self(locals()))

    @_return_requires(None)
    def volume_read_cache_policy_update(self, volume, rcp, flags=FLAG_RSVD):
        """
        lsm.Client.volume_read_cache_policy_update(self, volume, rcp,
                                                   flags=lsm.Client.FLAG_RSVD)

        Version:
            1.3
        Usage:
            Change the RAM read cache policy of specified volume.
            If lsm.Capabilities.VOLUME_READ_CACHE_POLICY_SET_IMPACT_WRITE
            is supported(like HPE SmartArray), the change on write cache policy
            might also impact read cache policy.
        Parameters:
            volume (Lsm.Volume)
                The lsm.Volume instance.
            rcp (int)
                Could be one of these value:
                    * lsm.Volume.READ_CACHE_POLICY_ENABLED
                        Enable read cache.
                    * lsm.Volume.READ_CACHE_POLICY_DISABLED
                        Disable read cache.
            flags (int, optional):
                Reserved for future use. Should be set as lsm.Client.FLAG_RSVD
        Returns:
            N/A
        SpecialExceptions:
            LsmError
                ErrorNumber.NO_SUPPORT
                ErrorNumber.NOT_FOUND_VOLUME
        Capability:
            lsm.Capabilities.VOLUME_READ_CACHE_POLICY_SET
                Allow enabling or disabling read cache policy.
            lsm.Capabilities.VOLUME_READ_CACHE_POLICY_SET_IMPACT_WRITE
                Indicate that changing read cache policy might impact write
                cache policy. For example, on HPE SmartArray, disabling read
                cache will also change write cache policy to write through.
        """
        if rcp != Volume.READ_CACHE_POLICY_ENABLED and \
           rcp != Volume.READ_CACHE_POLICY_DISABLED:
            raise LsmError(ErrorNumber.INVALID_ARGUMENT,
                           "Argument rcp should be "
                           "Volume.READ_CACHE_POLICY_ENABLED or "
                           "Volume.READ_CACHE_POLICY_DISABLED")
        return self._tp.rpc('volume_read_cache_policy_update',
                            _del_self(locals()))



# Copyright (C) 2018 Red Hat, Inc. Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin

from os import listdir
from os.path import isdir, join


class Fibrechannel(Plugin, RedHatPlugin):
    """Collects information on fibrechannel devices, if present"""

    plugin_name = 'fibrechannel'
    profiles = ('hardware', 'storage', 'system')
    files = ('/sys/class/fc_host')

    def setup(self):

        dirs = [
            '/sys/class/fc_host/',
            '/sys/class/fc_remote_ports/',
            '/sys/class/fc_transport/',
            '/sys/class/fc_vports/'
        ]

        devs = [join(d, dev) for d in dirs if isdir(d) for dev in listdir(d)]
        if devs:
            self.add_udev_info(devs, attrs=True)

# vim: set et ts=4 sw=4 :



# Copyright (c) 2014-2016 Hewlett Packard Enterprise Development LP
#
# Redistribution and use of this software in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

from requestbuilder import Arg, Filter, GenericTagFilter
import six

from euca2ools.commands.ec2 import EC2Request


class DescribeVpnConnections(EC2Request):
    DESCRIPTION = 'Show information about VPN connections'
    ARGS = [Arg('VpnConnectionId', metavar='VPNCONN', nargs='*',
                help='limit results to specific VPN connections'),
            Arg('--format', route_to=None, help='''show connection
                information in a specific format (cisco-ios-isr,
                juniper-junos-j, juniper-screenos-6.1, juniper-screenos-6.2,
                generic, xml, none) (default: none)'''),
            Arg('--stylesheet', route_to=None, help='''format the connection
                information using an XSL stylesheet.  If the value contains
                "{format}" it will be replaced with the format chosen by the
                --format option.  If the value is an HTTP or HTTPS URL it
                will be downloaded as needed.  (default: value of
                "vpn-stylesheet" region option)''')]
    FILTERS = [Filter('bgp-asn', help='''the BGP AS number advertised by the
                      customer gateway router'''),
               Filter('customer-gateway-configuration',
                      help='connection information for the customer gateway'),
               Filter('customer-gateway-id',
                      help='ID of the connected customer gateway'),
               Filter('state', help='''the VPN connection's state (available,
                      deleting, deleted, pending)'''),
               Filter('option.static-routes-only',
                      help='''whether the VPN connection is restricted to
                      static routes instead of using BGP'''),
               Filter('route.destination-cidr-block', help='''the address block
                      corresponding to the subnet used in the data center
                      behind the customer gateway router'''),
               Filter('tag-key',
                      help='key of a tag assigned to the VPN connection'),
               Filter('tag-value',
                      help='value of a tag assigned to the VPN connection'),
               GenericTagFilter('tag:KEY',
                                help='specific tag key/value combination'),
               Filter('type',
                      help='the type of virtual private gateway (ipsec.1)'),
               Filter('vpn-connection-id', help='ID of the VPN connection'),
               Filter('vpn-gateway-id',
                      help='ID of the connected virtual private gateway')]
    LIST_TAGS = ['vpnConnectionSet', 'tagSet']

    def print_result(self, result):
        if self.args.get('format') is None:
            stylesheet = self.args.get('stylesheet')
            show_conn_info = bool(stylesheet)
        elif self.args.get('format') == 'none':
            stylesheet = None
            show_conn_info = False
        elif self.args.get('format') == 'xml':
            stylesheet = None
            show_conn_info = True
        else:
            stylesheet = self.args.get('stylesheet')
            if not stylesheet:
                stylesheet = self.config.get_region_option('vpn-stylesheet')
            if stylesheet:
                stylesheet = stylesheet.format(format=self.args['format'])
            else:
                self.log.warn('current region has no stylesheet')
                msg = ('current region has no XSLT stylesheet to format '
                       'output; connection info will not be shown  (try '
                       'specifying one with "--stylesheet" or using '
                       '"--format xml")')
                six.print_(msg, file=sys.stderr)
            show_conn_info = bool(stylesheet)
        for vpn in result.get('vpnConnectionSet', []):
            self.print_vpn_connection(vpn, show_conn_info=show_conn_info,
                                      stylesheet=stylesheet)



# Copyright (C) 2018 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin
from pipes import quote
from re import match


class Candlepin(Plugin, RedHatPlugin):
    """Candlepin entitlement management"""

    plugin_name = 'candlepin'
    packages = ('candlepin',)

    def setup(self):
        # for external DB, search in /etc/candlepin/candlepin.conf for:
        # org.quartz.dataSource.myDS.URL=..
        #
        # and for DB password, search for
        # org.quartz.dataSource.myDS.password=..
        self.dbhost = "localhost"
        self.dbpasswd = ""
        cfg_file = "/etc/candlepin/candlepin.conf"
        try:
            for line in open(cfg_file).read().splitlines():
                # skip empty lines and lines with comments
                if not line or line[0] == '#':
                    continue
                if match(r"^\s*org.quartz.dataSource.myDS.URL=\S+", line):
                    self.dbhost = line.split('=')[1]
                    # separate hostname from value like
                    # jdbc:postgresql://localhost:5432/candlepin
                    self.dbhost = self.dbhost.split('/')[2].split(':')[0]
                if match(r"^\s*org.quartz.dataSource.myDS.password=\S+", line):
                    self.dbpasswd = line.split('=')[1]
        except (IOError, IndexError):
            # fallback when the cfg file is not accessible or parseable
            pass
        # set the password to os.environ when calling psql commands to prevent
        # printing it in sos logs
        # we can't set os.environ directly now: other plugins can overwrite it
        self.env = {"PGPASSWORD": self.dbpasswd}

        # Always collect the full active log of these
        self.add_copy_spec([
            "/var/log/candlepin/error.log",
            "/var/log/candlepin/candlepin.log"
        ], sizelimit=0)

        # Allow limiting on logrotated logs
        self.add_copy_spec([
            "/etc/candlepin/candlepin.conf",
            "/var/log/candlepin/audit*.log*",
            "/var/log/candlepin/candlepin.log[.-]*",
            "/var/log/candlepin/cpdb*.log*",
            "/var/log/candlepin/cpinit*.log*",
            "/var/log/candlepin/error.log[.-]*"
        ])

        self.add_cmd_output("du -sh /var/lib/candlepin/*/*")
        # collect tables sizes, ordered
        _cmd = self.build_query_cmd("\
            SELECT schema_name, relname, \
                   pg_size_pretty(table_size) AS size, table_size \
            FROM ( \
              SELECT \
                pg_catalog.pg_namespace.nspname AS schema_name, \
                relname, \
                pg_relation_size(pg_catalog.pg_class.oid) AS table_size \
              FROM pg_catalog.pg_class \
              JOIN pg_catalog.pg_namespace \
                ON relnamespace = pg_catalog.pg_namespace.oid \
            ) t \
            WHERE schema_name NOT LIKE 'pg_%' \
            ORDER BY table_size DESC;")
        self.add_cmd_output(_cmd, suggest_filename='candlepin_db_tables_sizes',
                            env=self.env)

	_cmd = self.build_query_cmd("\
            SELECT displayname, content_access_mode \
            FROM cp_owner;")
        self.add_cmd_output(_cmd,
                            suggest_filename='simple_content_access',
                            env=self.env)

    def build_query_cmd(self, query, csv=False):
        """
        Builds the command needed to invoke the pgsql query as the postgres
        user.
        The query requires significant quoting work to satisfy both the
        shell and postgres parsing requirements. Note that this will generate
        a large amount of quoting in sos logs referencing the command being run
        """
        csvformat = "-A -F , -X" if csv else ""
        _dbcmd = "psql --no-password -h %s -p 5432 -U candlepin \
                  -d candlepin %s -c %s"
        return _dbcmd % (self.dbhost, csvformat, quote(query))

    def postproc(self):
        reg = r"(((.*)(pass|token|secret)(.*))=)(.*)"
        repl = r"\1********"
        self.do_file_sub("/etc/candlepin/candlepin.conf", reg, repl)
        cpdbreg = r"(--password=)([a-zA-Z0-9]*)"
        self.do_file_sub("/var/log/candlepin/cpdb.log", cpdbreg, repl)

# vim: set et ts=4 sw=4 :



# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin, DebianPlugin, UbuntuPlugin


class OpenStackOctavia(Plugin):
    """Openstack Octavia"""

    plugin_name = "openstack_octavia"
    profiles = ('openstack', 'openstack_controller')

    var_puppet_gen = "/var/lib/config-data/puppet-generated/octavia"

    def setup(self):
        # configs
        self.add_copy_spec([
            "/etc/sysconfig/network-scripts/ifcfg-o-hm0",
            "/etc/logrotate.d/openstack-octavia",
            "/etc/octavia/*",
            "/var/lib/octavia",
            self.var_puppet_gen + "/etc/octavia/conf.d",
            self.var_puppet_gen + "/etc/octavia/octavia.conf",
            self.var_puppet_gen + "/etc/my.cnf.d/tripleo.cnf",
        ])

        # don't collect certificates
        self.add_forbidden_path("/etc/octavia/certs")

        # logs
        if self.get_option("all_logs"):
            self.add_copy_spec([
                "/var/log/octavia/*",
            ])
        else:
            self.add_copy_spec([
                "/var/log/octavia/*.log",
            ])

        # commands
        self.add_cmd_output([
            "openstack loadbalancer list",
            "openstack loadbalancer amphora list",
            "openstack loadbalancer healthmonitor list",
            "openstack loadbalancer l7policy list",
            "openstack loadbalancer listener list",
            "openstack loadbalancer pool list",
            "openstack loadbalancer quota list",
        ])

        # get details from each loadbalancer
        cmd = "openstack loadbalancer list -f value -c id"
        loadbalancers = self.exec_cmd(cmd)['output']
        for loadbalancer in loadbalancers.splitlines():
            loadbalancer = loadbalancer.split()[0]
            self.add_cmd_output(
                "openstack loadbalancer show %s" % (loadbalancer)
            )

        # get details from each l7policy
        cmd = "openstack loadbalancer l7policy list -f value -c id"
        l7policies = self.exec_cmd(cmd)['output']
        for l7policy in l7policies.splitlines():
            l7policy = l7policy.split()[0]
            self.add_cmd_output(
                "openstack loadbalancer l7rule list %s" % (l7policy)
            )

        # get details from each pool
        cmd = "openstack loadbalancer pool list -f value -c id"
        pools = self.exec_cmd(cmd)['output']
        for pool in pools.splitlines():
            pool = pool.split()[0]
            self.add_cmd_output(
                "openstack loadbalancer member list %s" % (pool)
            )

    def postproc(self):
        protect_keys = [
            "ca_private_key_passphrase", "heartbeat_key", "password",
            "connection", "transport_url"
        ]
        regexp = r"((?m)^\s*(%s)\s*=\s*)(.*)" % "|".join(protect_keys)

        self.do_path_regex_sub("/etc/octavia/*", regexp, r"\1*********")
        self.do_path_regex_sub(
            self.var_puppet_gen + "/etc/octavia/*",
            regexp, r"\1*********"
        )


class DebianOctavia(OpenStackOctavia, DebianPlugin, UbuntuPlugin):

    packages = ('octavia-common',)


class RedHatOctavia(OpenStackOctavia, RedHatPlugin):

    packages = ('openstack-selinux',)

# vim: set et ts=4 sw=4 :



# Copyright (C) 2017 Canonical Ltd.
#
# This file is part of cloud-init. See LICENSE file for license information.

"""Define 'devel' subcommand argument parsers to include in cloud-init cmd."""

import argparse
from cloudinit.config import schema

from . import net_convert
from . import render


def get_parser(parser=None):
    if not parser:
        parser = argparse.ArgumentParser(
            prog='cloudinit-devel',
            description='Run development cloud-init tools')
    subparsers = parser.add_subparsers(title='Subcommands', dest='subcommand')
    subparsers.required = True

    subcmds = [
        ('schema', 'Validate cloud-config files for document schema',
         schema.get_parser, schema.handle_schema_args),
        (net_convert.NAME, net_convert.__doc__,
         net_convert.get_parser, net_convert.handle_args),
        (render.NAME, render.__doc__,
         render.get_parser, render.handle_args)
    ]
    for (subcmd, helpmsg, get_parser, handler) in subcmds:
        parser = subparsers.add_parser(subcmd, help=helpmsg)
        get_parser(parser)
        parser.set_defaults(action=(subcmd, handler))

    return parser



# Copyright 2010 Red Hat, Inc.
# Author: Adam Stokes <astokes@fedoraproject.org>

# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.


"""
This module houses the i18n setup and message function. The default is to use
gettext to internationalize messages.
"""

import gettext
import six

from argparse import ArgumentParser

if six.PY3:
    from configparser import ConfigParser, ParsingError, Error
else:
    from ConfigParser import ConfigParser, ParsingError, Error

__version__ = "3.9"

gettext_dir = "/usr/share/locale"
gettext_app = "sos"

gettext.bindtextdomain(gettext_app, gettext_dir)


def _default(msg):
    return gettext.dgettext(gettext_app, msg)


_sos = _default

# Global option definitions
# These must be in the module itself in order to be available to both
# the sosreport and policy module (and to avoid recursive import errors).
#
# FIXME: these definitions make our main module a bit more bulky: the
# alternative is to place these in a new sos.options module. This may
# prove to be the best route long-term (as it could also contain an
# exported parsing routine, and all the command-line definitions).

#: Names of all arguments
_arg_names = [
    'add_preset', 'alloptions', 'allow_system_changes', 'all_logs', 'batch',
    'build', 'case_id', 'chroot', 'compression_type', 'config_file', 'desc',
    'debug', 'del_preset', 'dry_run', 'enableplugins', 'encrypt_key',
    'encrypt_pass', 'experimental', 'label', 'list_plugins', 'list_presets',
    'list_profiles', 'log_size', 'noplugins', 'noreport', 'no_env_vars',
    'no_postproc', 'note', 'onlyplugins', 'plugin_timeout', 'plugopts',
    'preset', 'profiles', 'quiet', 'since', 'sysroot', 'threads', 'tmp_dir',
    'upload', 'upload_url', 'upload_directory', 'upload_user', 'upload_pass',
    'upload_protocol', 'verbosity', 'verify'
]

#: Arguments with non-zero default values
_arg_defaults = {
    "chroot": "auto",
    "compression_type": "auto",
    "log_size": 25,
    "preset": "auto",
    # Verbosity has an explicit zero default since the ArgumentParser
    # count action default is None.
    "verbosity": 0,
    "upload_protocol": "auto"
}


def _is_seq(val):
    """Return true if val is an instance of a known sequence type.
    """
    val_type = type(val)
    return val_type is list or val_type is tuple


class SoSOptions(object):

    def _merge_opt(self, opt, src, is_default):
        def _unset(val):
            return (val == "" or val is None)

        if hasattr(src, opt):
            newvalue = getattr(src, opt)
            oldvalue = getattr(self, opt)
            # overwrite value iff:
            # - we replace unset option by a real value
            # - new default is set, or
            # - non-sequential variable keeps its default value
            if (_unset(oldvalue) and not _unset(newvalue)) or \
               is_default or \
               ((opt not in self._nondefault) and (not _is_seq(newvalue))):
                # Overwrite atomic values
                setattr(self, opt, newvalue)
                if is_default:
                    self._nondefault.discard(opt)
                else:
                    self._nondefault.add(opt)
            elif _is_seq(newvalue):
                # Concatenate sequence types
                setattr(self, opt, newvalue + oldvalue)

    def _merge_opts(self, src, is_default):
        for arg in _arg_names:
            self._merge_opt(arg, src, is_default)

    def __str(self, quote=False, sep=" ", prefix="", suffix=""):
        """Format a SoSOptions object as a human or machine readable string.

            :param quote: quote option values
            :param sep: list separator string
            :param prefix: arbitrary prefix string
            :param suffix: arbitrary suffix string
            :param literal: print values as Python literals
        """
        args = prefix
        arg_fmt = "=%s"
        for arg in _arg_names:
            args += arg + arg_fmt + sep
        args.strip(sep)

        vals = [getattr(self, arg) for arg in _arg_names]
        if not quote:
            # Convert Python source notation for sequences into plain strings
            vals = [",".join(v) if _is_seq(v) else v for v in vals]
        else:
            def is_string(val):
                return isinstance(val, six.string_types)
            # Only quote strings if quote=False
            vals = ["'%s'" % v if is_string(v) else v for v in vals]

        return (args % tuple(vals)).strip(sep) + suffix

    def __str__(self):
        return self.__str()

    def __repr__(self):
        return self.__str(quote=True, sep=", ", prefix="SoSOptions(",
                          suffix=")")

    def __init__(self, **kwargs):
        """Initialise a new ``SoSOptions`` object from keyword arguments.

            Initialises the new object with values taken from keyword
            arguments matching the names of ``SoSOptions`` attributes.

            A ``ValueError`` is raised is any of the supplied keyword
            arguments does not correspond to a known ``SoSOptions`
            attribute name.

            :param *kwargs: a list of ``SoSOptions`` keyword args.
            :returns: the new ``SoSOptions`` object.
        """
        self.add_preset = ""
        self.alloptions = False
        self.all_logs = False
        self.since = None
        self.batch = False
        self.build = False
        self.case_id = ""
        self.chroot = _arg_defaults["chroot"]
        self.compression_type = _arg_defaults["compression_type"]
        self.config_file = ""
        self.debug = False
        self.del_preset = ""
        self.desc = ""
        self.dry_run = False
        self.enableplugins = []
        self.encrypt_key = None
        self.encrypt_pass = None
        self.experimental = False
        self.label = ""
        self.list_plugins = False
        self.list_presets = False
        self.list_profiles = False
        self.log_size = _arg_defaults["log_size"]
        self.noplugins = []
        self.noreport = False
        self.allow_system_changes = False
        self.no_env_vars = False
        self.no_postproc = False
        self.note = ""
        self.onlyplugins = []
        self.plugin_timeout = None
        self.plugopts = []
        self.preset = _arg_defaults["preset"]
        self.profiles = []
        self.quiet = False
        self.sysroot = None
        self.threads = 4
        self.tmp_dir = ""
        self.upload = False
        self.upload_url = ""
        self.upload_directory = ""
        self.upload_user = ""
        self.upload_pass = ""
        self.upload_method = "auto"
        self.upload_protocol = _arg_defaults["upload_protocol"]
        self.verbosity = _arg_defaults["verbosity"]
        self.verify = False
        self._nondefault = set()
        for arg in kwargs.keys():
            if arg not in _arg_names:
                raise ValueError("Unknown SoSOptions attribute: %s" % arg)
            setattr(self, arg, kwargs[arg])

    @classmethod
    def from_args(cls, args):
        """Initialise a new SoSOptions object from a ``Namespace``
            obtained by parsing command line arguments.

            :param args: parsed command line arguments
            :returns: an initialised SoSOptions object
            :returntype: SoSOptions
        """
        opts = SoSOptions()
        opts._merge_opts(args, True)
        return opts

    @classmethod
    def _opt_to_args(cls, opt, val):
        """Convert a named option and optional value to command line
            argument notation, correctly handling options that take
            no value or that have special representations (e.g. verify
            and verbose).
        """
        no_value = (
            "alloptions", "allow-system-changes", "all-logs", "batch", "build",
            "debug", "experimental", "list-plugins", "list-presets",
            "list-profiles", "no-report", "no-env-vars", "quiet", "verify"
        )
        count = ("verbose",)
        if opt in no_value:
            return ["--%s" % opt]
        if opt in count:
            return ["--%s" % opt for d in range(0, int(val))]
        return ["--" + opt + "=" + val]

    @classmethod
    def from_file(cls, argparser, config_file, is_default=True):
        opts = SoSOptions()
        config = ConfigParser()
        try:
            try:
                with open(config_file) as f:
                    config.readfp(f)
            except (ParsingError, Error) as e:
                raise exit('Failed to parse configuration '
                           'file %s' % config_file)
        except (OSError, IOError) as e:
            raise exit('Unable to read configuration file %s '
                       ': %s' % (config_file, e.args[1]))

        if config.has_section("general"):
            optlist = []
            for opt, val in config.items("general"):
                optlist.extend(SoSOptions._opt_to_args(opt, val))
            opts._merge_opts(argparser.parse_args(optlist), is_default)

        opts.noplugins = []
        if config.has_option("plugins", "disable"):
            opts.noplugins.extend([plugin.strip() for plugin in
                                  config.get("plugins", "disable").split(',')])

        if config.has_option("plugins", "enable"):
            opts.enableplugins = []
            opts.enableplugins.extend(
                    [plugin.strip() for plugin in
                     config.get("plugins", "enable").split(',')])

        if config.has_section("tunables"):
            opts.plugopts = []
            for opt, val in config.items("tunables"):
                if not opt.split('.')[0] in opts.noplugins:
                    opts.plugopts.append(opt + "=" + val)

        return opts

    def merge(self, src, skip_default=True):
        """Merge another set of ``SoSOptions`` into this object.

            Merge two ``SoSOptions`` objects by setting unset or default
            values to their value in the ``src`` object.

            :param src: the ``SoSOptions`` object to copy from
            :param is_default: ``True`` if new default values are to be set.
        """
        for arg in _arg_names:
            if not hasattr(src, arg):
                continue
            if getattr(src, arg) is not None or not skip_default:
                self._merge_opt(arg, src, False)

    def dict(self):
        """Return this ``SoSOptions`` option values as a dictionary of
            argument name to value mappings.

            :returns: a name:value dictionary of option values.
        """
        odict = {}
        for arg in _arg_names:
            value = getattr(self, arg)
            # Do not attempt to store preset option values in presets
            if arg in ('add_preset', 'del_preset', 'desc', 'note'):
                value = None
            odict[arg] = value
        return odict

    def to_args(self):
        """Return command arguments for this object.

            Return a list of the non-default options of this ``SoSOptions``
            object in ``sosreport`` command line argument notation:

                ``["--all-logs", "-vvv"]``

        """
        def has_value(name, value):
            """ Test for non-null option values.
            """
            null_values = ("False", "None", "[]", '""', "''", "0")
            if not value or value in null_values:
                return False
            if name in _arg_defaults:
                if str(value) == str(_arg_defaults[name]):
                    return False
            return True

        def filter_opt(name, value):
            """ Filter out preset and null-valued options.
            """
            if name in ("add_preset", "del_preset", "desc", "note"):
                return False
            return has_value(name, value)

        def argify(name, value):
            """ Convert sos option notation to command line arguments.
            """
            # Handle --verbosity specially
            if name.startswith("verbosity"):
                arg = "-" + int(value) * "v"
                return arg

            name = name.replace("_", "-")

            value = ",".join(value) if _is_seq(value) else value

            if value is not True:
                opt = "%s %s" % (name, value)
            else:
                opt = name

            arg = "--" + opt if len(opt) > 1 else "-" + opt
            return arg

        opt_items = sorted(self.dict().items(), key=lambda x: x[0])
        return [argify(n, v) for (n, v) in opt_items if filter_opt(n, v)]

# vim: set et ts=4 sw=4 :



# code for cores isolation was inspired by Tuna implementation
# perf code was borrowed from kernel/tools/perf/python/twatch.py
# thanks to Arnaldo Carvalho de Melo <acme@redhat.com>

from . import base
from .decorators import *
import tuned.logs
import re
from subprocess import *
import threading
import perf
import select
import tuned.consts as consts
import procfs
import schedutils
from tuned.utils.commands import commands
import errno

log = tuned.logs.get()

class SchedulerParams(object):
	def __init__(self, cmd, cmdline = None, scheduler = None,
			priority = None, affinity = None):
		self._cmd = cmd
		self.cmdline = cmdline
		self.scheduler = scheduler
		self.priority = priority
		self.affinity = affinity

	@property
	def affinity(self):
		if self._affinity is None:
			return None
		else:
			return self._cmd.bitmask2cpulist(self._affinity)

	@affinity.setter
	def affinity(self, value):
		if value is None:
			self._affinity = None
		else:
			self._affinity = self._cmd.cpulist2bitmask(value)

class IRQAffinities(object):
	def __init__(self):
		self.irqs = {}
		self.default = None

class SchedulerPlugin(base.Plugin):
	"""
	Plugin for tuning of scheduler. Currently it can control scheduling
	priorities of system threads (it is substitution for the rtctl tool).
	"""

	_dict_schedcfg2num = {
			"f": schedutils.SCHED_FIFO,
			"b": schedutils.SCHED_BATCH,
			"r": schedutils.SCHED_RR,
			"o": schedutils.SCHED_OTHER,
			"i": schedutils.SCHED_IDLE,
			}

	def __init__(self, monitor_repository, storage_factory, hardware_inventory, device_matcher, device_matcher_udev, plugin_instance_factory, global_cfg, variables):
		super(SchedulerPlugin, self).__init__(monitor_repository, storage_factory, hardware_inventory, device_matcher, device_matcher_udev, plugin_instance_factory, global_cfg, variables)
		self._has_dynamic_options = True
		self._daemon = consts.CFG_DEF_DAEMON
		self._sleep_interval = int(consts.CFG_DEF_SLEEP_INTERVAL)
		if global_cfg is not None:
			self._daemon = global_cfg.get_bool(consts.CFG_DAEMON, consts.CFG_DEF_DAEMON)
			self._sleep_interval = int(global_cfg.get(consts.CFG_SLEEP_INTERVAL, consts.CFG_DEF_SLEEP_INTERVAL))
		self._cmd = commands()
		# default is to whitelist all and blacklist none
		self._ps_whitelist = ".*"
		self._ps_blacklist = ""
		self._cpus = perf.cpu_map()
		self._scheduler_storage_key = self._storage_key(
				command_name = "scheduler")
		self._irq_storage_key = self._storage_key(
				command_name = "irq")

	def _instance_init(self, instance):
		instance._has_dynamic_tuning = False
		instance._has_static_tuning = True
		# this is hack, runtime_tuning should be covered by dynamic_tuning configuration
		# TODO: add per plugin dynamic tuning configuration and use dynamic_tuning configuration
		# instead of runtime_tuning
		instance._runtime_tuning = True

		# FIXME: do we want to do this here?
		# recover original values in case of crash
		self._scheduler_original = self._storage.get(
				self._scheduler_storage_key, {})
		if len(self._scheduler_original) > 0:
			log.info("recovering scheduling settings from previous run")
			self._restore_ps_affinity()
			self._scheduler_original = {}
			self._storage.unset(self._scheduler_storage_key)

		instance._scheduler = instance.options
		for k in instance._scheduler:
			instance._scheduler[k] = self._variables.expand(instance._scheduler[k])
		if self._cmd.get_bool(instance._scheduler.get("runtime", 1)) == "0":
			instance._runtime_tuning = False
		instance._terminate = threading.Event()
		if self._daemon and instance._runtime_tuning:
			try:
				instance._threads = perf.thread_map()
				evsel = perf.evsel(type = perf.TYPE_SOFTWARE,
					config = perf.COUNT_SW_DUMMY,
					task = 1, comm = 1, mmap = 0, freq = 0,
					wakeup_events = 1, watermark = 1,
					sample_type = perf.SAMPLE_TID | perf.SAMPLE_CPU)
				evsel.open(cpus = self._cpus, threads = instance._threads)
				instance._evlist = perf.evlist(self._cpus, instance._threads)
				instance._evlist.add(evsel)
				instance._evlist.mmap()
			# no perf
			except:
				instance._runtime_tuning = False

	def _instance_cleanup(self, instance):
		pass

	@classmethod
	def _get_config_options(cls):
		return {
			"isolated_cores":       None,
			"ps_whitelist": None,
			"ps_blacklist": None,
		}

	# Raises OSError, IOError
	def _get_cmdline(self, process):
		if not isinstance(process, procfs.process):
			pid = process
			process = procfs.process(pid)
		cmdline = procfs.process_cmdline(process)
		if self._is_kthread(process):
			cmdline = "[" + cmdline + "]"
		return cmdline

	# Raises OSError, IOError
	def get_processes(self):
		ps = procfs.pidstats()
		ps.reload_threads()
		processes = {}
		for proc in ps.values():
			try:
				cmd = self._get_cmdline(proc)
				pid = proc["pid"]
				processes[pid] = cmd
				if "threads" in proc:
					for pid in proc["threads"].keys():
						cmd = self._get_cmdline(proc)
						processes[pid] = cmd
			except (OSError, IOError) as e:
				if e.errno == errno.ENOENT \
						or e.errno == errno.ESRCH:
					continue
				else:
					raise
		return processes

	# Raises OSError
	# Raises SystemError with old (pre-0.4) python-schedutils
	# instead of OSError
	# If PID doesn't exist, errno == ESRCH
	def _get_rt(self, pid):
		scheduler = schedutils.get_scheduler(pid)
		sched_str = schedutils.schedstr(scheduler)
		priority = schedutils.get_priority(pid)
		log.debug("Read scheduler policy '%s' and priority '%d' of PID '%d'"
				% (sched_str, priority, pid))
		return (scheduler, priority)

	def _set_rt(self, pid, sched, prio):
		sched_str = schedutils.schedstr(sched)
		log.debug("Setting scheduler policy to '%s' and priority to '%d' of PID '%d'."
				% (sched_str, prio, pid))
		try:
			prio_min = schedutils.get_priority_min(sched)
			prio_max = schedutils.get_priority_max(sched)
			if prio < prio_min or prio > prio_max:
				log.error("Priority for %s must be in range %d - %d. '%d' was given."
						% (sched_str, prio_min,
						prio_max, prio))
		# Workaround for old (pre-0.4) python-schedutils which raised
		# SystemError instead of OSError
		except (SystemError, OSError) as e:
			log.error("Failed to get allowed priority range: %s"
					% e)
		try:
			schedutils.set_scheduler(pid, sched, prio)
		except (SystemError, OSError) as e:
			if hasattr(e, "errno") and e.errno == errno.ESRCH:
				log.debug("Failed to set scheduling parameters of PID %d, the task vanished."
						% pid)
			else:
				log.error("Failed to set scheduling parameters of PID %d: %s"
						% (pid, e))

	# process is a procfs.process object
	# Raises OSError, IOError
	def _is_kthread(self, process):
		return process["stat"]["flags"] & procfs.pidstat.PF_KTHREAD != 0

	# Return codes:
	# 0 - Affinity is fixed
	# 1 - Affinity is changeable
	# -1 - Task vanished
	# -2 - Error
	def _affinity_changeable(self, pid):
		try:
			process = procfs.process(pid)
			if process["stat"].is_bound_to_cpu():
				if process["stat"]["state"] == "Z":
					log.debug("Affinity of zombie task with PID %d cannot be changed, the task's affinity mask is fixed."
							% pid)
				elif self._is_kthread(process):
					log.debug("Affinity of kernel thread with PID %d cannot be changed, the task's affinity mask is fixed."
							% pid)
				else:
					log.warn("Affinity of task with PID %d cannot be changed, the task's affinity mask is fixed."
							% pid)
				return 0
			else:
				return 1
		except (OSError, IOError) as e:
			if e.errno == errno.ENOENT or e.errno == errno.ESRCH:
				log.debug("Failed to get task info for PID %d, the task vanished."
						% pid)
				return -1
			else:
				log.error("Failed to get task info for PID %d: %s"
						% (pid, e))
				return -2
		except (AttributeError, KeyError) as e:
			log.error("Failed to get task info for PID %d: %s"
					% (pid, e))
			return -2

	def _store_orig_process_rt(self, pid, scheduler, priority):
		try:
			params = self._scheduler_original[pid]
		except KeyError:
			params = SchedulerParams(self._cmd)
			self._scheduler_original[pid] = params
		if params.scheduler is None and params.priority is None:
			params.scheduler = scheduler
			params.priority = priority

	def _tune_process_rt(self, pid, sched, prio):
		cont = True
		if sched is None and prio is None:
			return cont
		try:
			(prev_sched, prev_prio) = self._get_rt(pid)
			if sched is None:
				sched = prev_sched
			self._set_rt(pid, sched, prio)
			self._store_orig_process_rt(pid, prev_sched, prev_prio)
		except (SystemError, OSError) as e:
			if hasattr(e, "errno") and e.errno == errno.ESRCH:
				log.debug("Failed to read scheduler policy of PID %d, the task vanished."
						% pid)
				if pid in self._scheduler_original:
					del self._scheduler_original[pid]
				cont = False
			else:
				log.error("Refusing to set scheduler and priority of PID %d, reading original scheduling parameters failed: %s"
						% (pid, e))
		return cont

	def _store_orig_process_affinity(self, pid, affinity):
		try:
			params = self._scheduler_original[pid]
		except KeyError:
			params = SchedulerParams(self._cmd)
			self._scheduler_original[pid] = params
		if params.affinity is None:
			params.affinity = affinity

	def _tune_process_affinity(self, pid, affinity, intersect = False):
		cont = True
		if affinity is None:
			return cont
		try:
			prev_affinity = self._get_affinity(pid)
			if intersect:
				affinity = self._get_intersect_affinity(
						prev_affinity, affinity,
						affinity)
			self._set_affinity(pid, affinity)
			self._store_orig_process_affinity(pid,
					prev_affinity)
		except (SystemError, OSError) as e:
			if hasattr(e, "errno") and e.errno == errno.ESRCH:
				log.debug("Failed to read affinity of PID %d, the task vanished."
						% pid)
				if pid in self._scheduler_original:
					del self._scheduler_original[pid]
				cont = False
			else:
				log.error("Refusing to set CPU affinity of PID %d, reading original affinity failed: %s"
						% (pid, e))
		return cont

	#tune process and store previous values
	def _tune_process(self, pid, cmd, sched, prio, affinity):
		cont = self._tune_process_rt(pid, sched, prio)
		if not cont:
			return
		cont = self._tune_process_affinity(pid, affinity)
		if not cont or pid not in self._scheduler_original:
			return
		self._scheduler_original[pid].cmdline = cmd

	def _convert_sched_params(self, str_scheduler, str_priority):
		scheduler = self._dict_schedcfg2num.get(str_scheduler)
		if scheduler is None and str_scheduler != "*":
			log.error("Invalid scheduler: %s. Scheduler and priority will be ignored."
					% str_scheduler)
			return (None, None)
		else:
			try:
				priority = int(str_priority)
			except ValueError:
				log.error("Invalid priority: %s. Scheduler and priority will be ignored."
							% str_priority)
				return (None, None)
		return (scheduler, priority)

	def _convert_affinity(self, str_affinity):
		if str_affinity == "*":
			affinity = None
		else:
			affinity = self._cmd.hex2cpulist(str_affinity)
			if not affinity:
				log.error("Invalid affinity: %s. It will be ignored."
						% str_affinity)
				affinity = None
		return affinity

	def _convert_sched_cfg(self, vals):
		(rule_prio, scheduler, priority, affinity, regex) = vals
		(scheduler, priority) = self._convert_sched_params(
				scheduler, priority)
		affinity = self._convert_affinity(affinity)
		return (rule_prio, scheduler, priority, affinity, regex)

	def _instance_apply_static(self, instance):
		super(SchedulerPlugin, self)._instance_apply_static(instance)
		try:
			ps = self.get_processes()
		except (OSError, IOError) as e:
			log.error("error applying tuning, cannot get information about running processes: %s"
					% e)
			return
		sched_cfg = [(option, str(value).split(":", 4)) for option, value in instance._scheduler.items()]
		buf = [(option, self._convert_sched_cfg(vals))
				for option, vals in sched_cfg
				if re.match(r"group\.", option)
				and len(vals) == 5]
		sched_cfg = sorted(buf, key=lambda option_vals: option_vals[1][0])
		sched_all = dict()
		# for runtime tunning
		instance._sched_lookup = {}
		for option, (rule_prio, scheduler, priority, affinity, regex) \
				in sched_cfg:
			try:
				r = re.compile(regex)
			except re.error as e:
				log.error("error compiling regular expression: '%s'" % str(regex))
				continue
			processes = [(pid, cmd) for pid, cmd in ps.items() if re.search(r, cmd) is not None]
			#cmd - process name, option - group name
			sched = dict([(pid, (cmd, option, scheduler, priority, affinity, regex))
					for pid, cmd in processes])
			sched_all.update(sched)
			regex = str(regex).replace("(", r"\(")
			regex = regex.replace(")", r"\)")
			instance._sched_lookup[regex] = [scheduler, priority, affinity]
		for pid, (cmd, option, scheduler, priority, affinity, regex) \
				in sched_all.items():
			self._tune_process(pid, cmd, scheduler,
					priority, affinity)
		self._storage.set(self._scheduler_storage_key,
				self._scheduler_original)
		if self._daemon and instance._runtime_tuning:
			instance._thread = threading.Thread(target = self._thread_code, args = [instance])
			instance._thread.start()

	def _restore_ps_affinity(self):
		try:
			ps = self.get_processes()
		except (OSError, IOError) as e:
			log.error("error unapplying tuning, cannot get information about running processes: %s"
					% e)
			return
		for pid, orig_params in self._scheduler_original.items():
			# if command line for the pid didn't change, it's very probably the same process
			if pid not in ps or ps[pid] != orig_params.cmdline:
				continue
			if orig_params.scheduler is not None \
					and orig_params.priority is not None:
				self._set_rt(pid, orig_params.scheduler,
						orig_params.priority)
			if orig_params.affinity is not None:
				self._set_affinity(pid, orig_params.affinity)
		self._scheduler_original = {}
		self._storage.unset(self._scheduler_storage_key)

	def _instance_unapply_static(self, instance, full_rollback = False):
		super(SchedulerPlugin, self)._instance_unapply_static(instance, full_rollback)
		if self._daemon and instance._runtime_tuning:
			instance._terminate.set()
			instance._thread.join()
		self._restore_ps_affinity()

	def _add_pid(self, instance, pid, r):
		try:
			cmd = self._get_cmdline(pid)
		except (OSError, IOError) as e:
			if e.errno == errno.ENOENT \
					or e.errno == errno.ESRCH:
				log.debug("Failed to get cmdline of PID %d, the task vanished."
						% pid)
			else:
				log.error("Failed to get cmdline of PID %d: %s"
						% (pid, e))
			return
		v = self._cmd.re_lookup(instance._sched_lookup, cmd, r)
		if v is not None and not pid in self._scheduler_original:
			log.debug("tuning new process '%s' with PID '%d' by '%s'" % (cmd, pid, str(v)))
			(sched, prio, affinity) = v
			self._tune_process(pid, cmd, sched, prio,
					affinity)
			self._storage.set(self._scheduler_storage_key,
					self._scheduler_original)

	def _remove_pid(self, instance, pid):
		if pid in self._scheduler_original:
			del self._scheduler_original[pid]
			log.debug("removed PID %d from the rollback database" % pid)
			self._storage.set(self._scheduler_storage_key,
					self._scheduler_original)

	def _thread_code(self, instance):
		r = self._cmd.re_lookup_compile(instance._sched_lookup)
		poll = select.poll()
		# Store the file objects in a local variable so that they don't
		# go out of scope too soon. This is a workaround for
		# python3-perf bug rhbz#1659445.
		fds = instance._evlist.get_pollfd()
		for fd in fds:
			poll.register(fd)
		while not instance._terminate.is_set():
			# timeout to poll in milliseconds
			if len(poll.poll(self._sleep_interval * 1000)) > 0 and not instance._terminate.is_set():
				read_events = True
				while read_events:
					read_events = False
					for cpu in self._cpus:
						event = instance._evlist.read_on_cpu(cpu)
						if event:
							read_events = True
							if event.type == perf.RECORD_COMM:
								self._add_pid(instance, int(event.tid), r)
							elif event.type == perf.RECORD_EXIT:
								self._remove_pid(instance, int(event.tid))

	@command_custom("ps_whitelist", per_device = False)
	def _ps_whitelist(self, enabling, value, verify, ignore_missing):
		# currently unsupported
		if verify:
			return None
		if enabling and value is not None:
			self._ps_whitelist = "|".join(["(%s)" % v for v in re.split(r"(?<!\\);", str(value))])

	@command_custom("ps_blacklist", per_device = False)
	def _ps_blacklist(self, enabling, value, verify, ignore_missing):
		# currently unsupported
		if verify:
			return None
		if enabling and value is not None:
			self._ps_blacklist = "|".join(["(%s)" % v for v in re.split(r"(?<!\\);", str(value))])

	# Raises OSError
	# Raises SystemError with old (pre-0.4) python-schedutils
	# instead of OSError
	# If PID doesn't exist, errno == ESRCH
	def _get_affinity(self, pid):
		res = schedutils.get_affinity(pid)
		log.debug("Read affinity '%s' of PID %d" % (res, pid))
		return res

	def _set_affinity(self, pid, affinity):
		log.debug("Setting CPU affinity of PID %d to '%s'." % (pid, affinity))
		try:
			schedutils.set_affinity(pid, affinity)
			return True
		# Workaround for old python-schedutils (pre-0.4) which
		# incorrectly raised SystemError instead of OSError
		except (SystemError, OSError) as e:
			if hasattr(e, "errno") and e.errno == errno.ESRCH:
				log.debug("Failed to set affinity of PID %d, the task vanished."
						% pid)
			else:
				res = self._affinity_changeable(pid)
				if res == 1 or res == -2:
					log.error("Failed to set affinity of PID %d to '%s': %s"
							% (pid, affinity, e))
			return False

	# returns intersection of affinity1 with affinity2, if intersection is empty it returns affinity3
	def _get_intersect_affinity(self, affinity1, affinity2, affinity3):
		aff = set(affinity1).intersection(set(affinity2))
		if aff:
			return list(aff)
		return affinity3

	def _set_all_obj_affinity(self, objs, affinity, threads = False):
		psl = [v for v in objs if re.search(self._ps_whitelist,
				self._get_stat_comm(v)) is not None]
		if self._ps_blacklist != "":
			psl = [v for v in psl if re.search(self._ps_blacklist,
					self._get_stat_comm(v)) is None]
		psd = dict([(v.pid, v) for v in psl])
		for pid in psd:
			try:
				cmd = self._get_cmdline(psd[pid])
			except (OSError, IOError) as e:
				if e.errno == errno.ENOENT \
						or e.errno == errno.ESRCH:
					log.debug("Failed to get cmdline of PID %d, the task vanished."
							% pid)
				else:
					log.error("Refusing to set affinity of PID %d, failed to get its cmdline: %s"
							% (pid, e))
				continue
			cont = self._tune_process_affinity(pid, affinity,
					intersect = True)
			if not cont:
				continue
			if pid in self._scheduler_original:
				self._scheduler_original[pid].cmdline = cmd
			# process threads
			if not threads and "threads" in psd[pid]:
				self._set_all_obj_affinity(
						psd[pid]["threads"].values(),
						affinity, True)

	def _get_stat_comm(self, o):
		try:
			return o["stat"]["comm"]
		except (OSError, IOError, KeyError):
			return ""

	def _set_ps_affinity(self, affinity):
		try:
			ps = procfs.pidstats()
			ps.reload_threads()
			self._set_all_obj_affinity(ps.values(), affinity, False)
		except (OSError, IOError) as e:
			log.error("error applying tuning, cannot get information about running processes: %s"
					% e)

	def _set_irq_affinity(self, irq, affinity, restoring):
		try:
			affinity_hex = self._cmd.cpulist2hex(affinity)
			log.debug("Setting SMP affinity of IRQ %s to '%s'"
					% (irq, affinity_hex))
			filename = "/proc/irq/%s/smp_affinity" % irq
			with open(filename, "w") as f:
				f.write(affinity_hex)
			return True
		except (OSError, IOError) as e:
			# EIO is returned by
			# kernel/irq/proc.c:write_irq_affinity() if changing
			# the affinity is not supported
			# (at least on kernels 3.10 and 4.18)
			if hasattr(e, "errno") and e.errno == errno.EIO \
					and not restoring:
				log.debug("Setting SMP affinity of IRQ %s is not supported"
						% irq)
			else:
				log.error("Failed to set SMP affinity of IRQ %s to '%s': %s"
						% (irq, affinity_hex, e))
			return False

	def _set_default_irq_affinity(self, affinity):
		try:
			affinity_hex = self._cmd.cpulist2hex(affinity)
			log.debug("Setting default SMP IRQ affinity to '%s'"
					% affinity_hex)
			with open("/proc/irq/default_smp_affinity", "w") as f:
				f.write(affinity_hex)
		except (OSError, IOError) as e:
			log.error("Failed to set default SMP IRQ affinity to '%s': %s"
					% (affinity_hex, e))

	def _set_all_irq_affinity(self, affinity):
		irq_original = IRQAffinities()
		irqs = procfs.interrupts()
		for irq in irqs.keys():
			try:
				prev_affinity = irqs[irq]["affinity"]
				log.debug("Read affinity of IRQ '%s': '%s'"
						% (irq, prev_affinity))
			except KeyError:
				continue
			_affinity = self._get_intersect_affinity(prev_affinity, affinity, affinity)
			if set(_affinity) == set(prev_affinity):
				continue
			if self._set_irq_affinity(irq, _affinity, False):
				irq_original.irqs[irq] = prev_affinity

		# default affinity
		prev_affinity_hex = self._cmd.read_file("/proc/irq/default_smp_affinity")
		prev_affinity = self._cmd.hex2cpulist(prev_affinity_hex)
		_affinity = self._get_intersect_affinity(prev_affinity, affinity, affinity)
		self._set_default_irq_affinity(_affinity)
		irq_original.default = prev_affinity
		self._storage.set(self._irq_storage_key, irq_original)

	def _restore_all_irq_affinity(self):
		irq_original = self._storage.get(self._irq_storage_key, None)
		if irq_original is None:
			return
		for irq, affinity in irq_original.irqs.items():
			self._set_irq_affinity(irq, affinity, True)
		affinity = irq_original.default
		self._set_default_irq_affinity(affinity)
		self._storage.unset(self._irq_storage_key)

	def _verify_irq_affinity(self, irq_description, correct_affinity,
			current_affinity):
		res = set(current_affinity).issubset(set(correct_affinity))
		if res:
			log.info(consts.STR_VERIFY_PROFILE_VALUE_OK
					% (irq_description, current_affinity))
		else:
			log.error(consts.STR_VERIFY_PROFILE_VALUE_FAIL
					% (irq_description, current_affinity,
					correct_affinity))
		return res

	def _verify_all_irq_affinity(self, correct_affinity):
		irqs = procfs.interrupts()
		res = True
		for irq in irqs.keys():
			try:
				current_affinity = irqs[irq]["affinity"]
				log.debug("Read SMP affinity of IRQ '%s': '%s'"
						% (irq, current_affinity))
				irq_description = "SMP affinity of IRQ %s" % irq
				if not self._verify_irq_affinity(
						irq_description,
						correct_affinity,
						current_affinity):
					res = False
			except KeyError:
				continue

		current_affinity_hex = self._cmd.read_file(
				"/proc/irq/default_smp_affinity")
		current_affinity = self._cmd.hex2cpulist(current_affinity_hex)
		if not self._verify_irq_affinity("default IRQ SMP affinity",
				current_affinity, correct_affinity):
			res = False
		return res

	@command_custom("isolated_cores", per_device = False, priority = 10)
	def _isolated_cores(self, enabling, value, verify, ignore_missing):
		affinity = None
		if value is not None:
			isolated = set(self._cmd.cpulist_unpack(value))
			present = set(self._cpus)
			if isolated.issubset(present):
				affinity = list(present - isolated)
			else:
				str_cpus = ",".join([str(x) for x in self._cpus])
				log.error("Invalid isolated_cores specified, '%s' does not match available cores '%s'"
						% (value, str_cpus))
		if (enabling or verify) and affinity is None:
			return None
		# currently only IRQ affinity verification is supported
		if verify:
			return self._verify_all_irq_affinity(affinity)
		elif enabling:
			self._set_ps_affinity(affinity)
			self._set_all_irq_affinity(affinity)
		else:
			# Restoring processes' affinity is done in
			# _instance_unapply_static()
			self._restore_all_irq_affinity()



import tuned.logs
from configobj import ConfigObj, ConfigObjError
from validate import Validator
from tuned.exceptions import TunedException
import tuned.consts as consts
from tuned.utils.commands import commands

__all__ = ["GlobalConfig"]

log = tuned.logs.get()

class GlobalConfig():

	global_config_spec = ["dynamic_tuning = boolean(default=%s)" % consts.CFG_DEF_DYNAMIC_TUNING,
		"sleep_interval = integer(default=%s)" % consts.CFG_DEF_SLEEP_INTERVAL,
		"update_interval = integer(default=%s)" % consts.CFG_DEF_UPDATE_INTERVAL,
		"recommend_command = boolean(default=%s)" % consts.CFG_DEF_RECOMMEND_COMMAND]

	def __init__(self,config_file = consts.GLOBAL_CONFIG_FILE):
		self._cfg = {}
		self.load_config(file_name=config_file)
		self._cmd = commands()

	def load_config(self, file_name = consts.GLOBAL_CONFIG_FILE):
		"""
		Loads global configuration file.
		"""
		log.debug("reading and parsing global configuration file '%s'" % file_name)
		try:
			self._cfg = ConfigObj(file_name, configspec = self.global_config_spec, raise_errors = True, \
				file_error = True, list_values = False, interpolation = False)
		except IOError as e:
			raise TunedException("Global tuned configuration file '%s' not found." % file_name)
		except ConfigObjError as e:
			raise TunedException("Error parsing global tuned configuration file '%s'." % file_name)
		vdt = Validator()
		if (not self._cfg.validate(vdt, copy=True)):
			raise TunedException("Global tuned configuration file '%s' is not valid." % file_name)

	def get(self, key, default = None):
		return self._cfg.get(key, default)

	def get_bool(self, key, default = None):
		if self._cmd.get_bool(self.get(key, default)) == "1":
			return True
		return False

	def set(self, key, value):
		self._cfg[key] = value

	def get_size(self, key, default = None):
		val = self.get(key)
		if val is None:
			return default
		ret = self._cmd.get_size(val)
		if ret is None:
			log.error("Error parsing value '%s', using '%s'." %(val, default))
			return default
		else:
			return ret



# -*- coding: utf-8 -*-
#
# python-json-pointer - An implementation of the JSON Pointer syntax
# https://github.com/stefankoegl/python-json-pointer
#
# Copyright (c) 2011 Stefan Kögl <stefan@skoegl.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import unicode_literals

""" Identify specific nodes in a JSON document (RFC 6901) """

try:
    from collections.abc import Mapping, Sequence
except ImportError:
    from collections import Mapping, Sequence

# Will be parsed by setup.py to determine package metadata
__author__ = 'Stefan Kögl <stefan@skoegl.net>'
__version__ = '1.9'
__website__ = 'https://github.com/stefankoegl/python-json-pointer'
__license__ = 'Modified BSD License'


try:
    from urllib import unquote
    from itertools import izip
    str = unicode
except ImportError: # Python 3
    from urllib.parse import unquote
    izip = zip

from itertools import tee
import re
import copy


# array indices must not contain leading zeros, signs, spaces, decimals, etc
RE_ARRAY_INDEX=re.compile('0|[1-9][0-9]*$')


class JsonPointerException(Exception):
    pass


class EndOfList(object):
    """ Result of accessing element "-" of a list """

    def __init__(self, list_):
        self.list_ = list_


    def __repr__(self):
        return '{cls}({lst})'.format(cls=self.__class__.__name__,
                lst=repr(self.list_))


_nothing = object()


def resolve_pointer(doc, pointer, default=_nothing):
    """
    Resolves pointer against doc and returns the referenced object

    >>> obj = {"foo": {"anArray": [ {"prop": 44}], "another prop": {"baz": "A string" }}}

    >>> resolve_pointer(obj, '') == obj
    True

    >>> resolve_pointer(obj, '/foo') == obj['foo']
    True

    >>> resolve_pointer(obj, '/foo/another%20prop') == obj['foo']['another prop']
    True

    >>> resolve_pointer(obj, '/foo/another%20prop/baz') == obj['foo']['another prop']['baz']
    True

    >>> resolve_pointer(obj, '/foo/anArray/0') == obj['foo']['anArray'][0]
    True

    >>> resolve_pointer(obj, '/some/path', None) == None
    True

    """

    pointer = JsonPointer(pointer)
    return pointer.resolve(doc, default)

def set_pointer(doc, pointer, value, inplace=True):
    """
    Resolves pointer against doc and sets the value of the target within doc.

    With inplace set to true, doc is modified as long as pointer is not the
    root.

    >>> obj = {"foo": {"anArray": [ {"prop": 44}], "another prop": {"baz": "A string" }}}

    >>> set_pointer(obj, '/foo/anArray/0/prop', 55) == \
    {'foo': {'another prop': {'baz': 'A string'}, 'anArray': [{'prop': 55}]}}
    True

    >>> set_pointer(obj, '/foo/yet%20another%20prop', 'added prop') == \
    {'foo': {'another prop': {'baz': 'A string'}, 'yet another prop': 'added prop', 'anArray': [{'prop': 55}]}}
    True

    """

    pointer = JsonPointer(pointer)
    return pointer.set(doc, value, inplace)


class JsonPointer(object):
    """ A JSON Pointer that can reference parts of an JSON document """

    def __init__(self, pointer):
        parts = pointer.split('/')
        if parts.pop(0) != '':
            raise JsonPointerException('location must starts with /')

        parts = map(unquote, parts)
        parts = [part.replace('~1', '/') for part in parts]
        parts = [part.replace('~0', '~') for part in parts]
        self.parts = parts


    def to_last(self, doc):
        """ Resolves ptr until the last step, returns (sub-doc, last-step) """

        if not self.parts:
            return doc, None

        for part in self.parts[:-1]:
            doc = self.walk(doc, part)

        return doc, self.get_part(doc, self.parts[-1])


    def resolve(self, doc, default=_nothing):
        """Resolves the pointer against doc and returns the referenced object"""

        for part in self.parts:

            try:
                doc = self.walk(doc, part)
            except JsonPointerException:
                if default is _nothing:
                    raise
                else:
                    return default

        return doc


    get = resolve

    def set(self, doc, value, inplace=True):
        """ Resolve the pointer against the doc and replace the target with value. """

        if len(self.parts) == 0:
            if inplace:
                raise JsonPointerException('cannot set root in place')
            return value

        if not inplace:
            doc = copy.deepcopy(doc)

        (parent, part) = self.to_last(doc)

        parent[part] = value
        return doc

    def get_part(self, doc, part):
        """ Returns the next step in the correct type """

        if isinstance(doc, Mapping):
            return part

        elif isinstance(doc, Sequence):

            if part == '-':
                return part

            if not RE_ARRAY_INDEX.match(str(part)):
                raise JsonPointerException("'%s' is not a valid list index" % (part, ))

            return int(part)

        elif hasattr(doc, '__getitem__'):
            # Allow indexing via ducktyping if the target has defined __getitem__
            return part

        else:
            raise JsonPointerException("Document '%s' does not support indexing, "
                                       "must be dict/list or support __getitem__" % type(doc))


    def walk(self, doc, part):
        """ Walks one step in doc and returns the referenced part """

        part = self.get_part(doc, part)

        assert (type(doc) in (dict, list) or hasattr(doc, '__getitem__')), "invalid document type %s" % (type(doc),)

        if isinstance(doc, Mapping):
            try:
                return doc[part]

            except KeyError:
                raise JsonPointerException("member '%s' not found in %s" % (part, doc))

        elif isinstance(doc, Sequence):

            if part == '-':
                return EndOfList(doc)

            try:
                return doc[part]

            except IndexError:
                raise JsonPointerException("index '%s' is out of bounds" % (part, ))

        else:
            # Object supports __getitem__, assume custom indexing
            return doc[part]

    def contains(self, ptr):
        """Returns True if self contains the given ptr"""
        return self.parts[:len(ptr.parts)] == ptr.parts

    def __contains__(self, item):
        """Returns True if self contains the given ptr"""
        return self.contains(item)

    @property
    def path(self):
        """ Returns the string representation of the pointer

        >>> ptr = JsonPointer('/~0/0/~1').path == '/~0/0/~1'
        """
        parts = [part.replace('~', '~0') for part in self.parts]
        parts = [part.replace('/', '~1') for part in parts]
        return ''.join('/' + part for part in parts)

    def __eq__(self, other):
        """ compares a pointer to another object

        Pointers can be compared by comparing their strings (or splitted
        strings), because no two different parts can point to the same
        structure in an object (eg no different number representations) """

        if not isinstance(other, JsonPointer):
            return False

        return self.parts == other.parts


    def __hash__(self):
        return hash(tuple(self.parts))

    @classmethod
    def from_parts(cls, parts):
        """ Constructs a JsonPointer from a list of (unescaped) paths

        >>> JsonPointer.from_parts(['a', '~', '/', 0]).path == '/a/~0/~1/0'
        True
        """
        parts = [str(part) for part in parts]
        parts = [part.replace('~', '~0') for part in parts]
        parts = [part.replace('/', '~1') for part in parts]
        ptr = cls(''.join('/' + part for part in parts))
        return ptr



def pairwise(iterable):
    """ s -> (s0,s1), (s1,s2), (s2, s3), ...

    >>> list(pairwise([]))
    []

    >>> list(pairwise([1]))
    []

    >>> list(pairwise([1, 2, 3, 4]))
    [(1, 2), (2, 3), (3, 4)]
    """
    a, b = tee(iterable)
    for _ in b:
        break
    return izip(a, b)



# Copyright (C) 2018 Mark Michelson <mmichels@redhat.com>

# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin, DebianPlugin, UbuntuPlugin
import json
import os
import six


class OVNCentral(Plugin):
    """ OVN Northd
    """
    plugin_name = "ovn_central"
    profiles = ('network', 'virt')
    _container_runtime = None
    _container_name = None

    def get_tables_from_schema(self, filename, skip=[]):
        if self._container_name:
            cmd = "%s exec %s cat %s" % (
                self._container_runtime, self._container_name, filename)
            res = self.exec_cmd(cmd, foreground=True)
            if res['status'] != 0:
                self._log_error("Could not retrieve DB schema file from "
                                "container %s" % self._container_name)
                return
            try:
                db = json.loads(res['output'])
            except Exception:
                self._log_error("Cannot parse JSON file %s" % filename)
                return
        else:
            try:
                with open(filename, 'r') as f:
                    try:
                        db = json.load(f)
                    except Exception:
                        self._log_error(
                            "Cannot parse JSON file %s" % filename)
                        return
            except IOError as ex:
                self._log_error(
                    "Could not open DB schema file %s: %s" % (filename, ex))
                return
        try:
            return [table for table in six.iterkeys(
                db['tables']) if table not in skip]
        except AttributeError:
            self._log_error("DB schema %s has no 'tables' key" % filename)

    def add_database_output(self, tables, cmds, ovn_cmd):
        if not tables:
            return
        for table in tables:
            cmds.append('%s list %s' % (ovn_cmd, table))

    def running_in_container(self):
        for runtime in ["podman", "docker"]:
            container_status = self.exec_cmd(runtime + " ps")
            if container_status['status'] == 0:
                for line in container_status['output'].splitlines():
                    if "ovn-dbs-bundle" in line:
                        self._container_name = line.split()[-1]
                        self._container_runtime = runtime
                        return True
        return False

    def check_enabled(self):
        return (self.running_in_container() or
                super(OVNCentral, self).check_enabled())

    def setup(self):
        containerized = self.running_in_container()

        ovs_rundir = os.environ.get('OVS_RUNDIR')
        for pidfile in ['ovnnb_db.pid', 'ovnsb_db.pid', 'ovn-northd.pid']:
            self.add_copy_spec([
                os.path.join('/var/lib/openvswitch/ovn', pidfile),
                os.path.join('/usr/local/var/run/openvswitch', pidfile),
                os.path.join('/run/openvswitch/', pidfile),
            ])

            if ovs_rundir:
                self.add_copy_spec(os.path.join(ovs_rundir, pidfile))

        # Some user-friendly versions of DB output
        cmds = [
            'ovn-nbctl show',
            'ovn-sbctl show',
            'ovn-sbctl lflow-list',
            'ovn-nbctl get-ssl',
            'ovn-nbctl get-connection',
            'ovn-sbctl get-ssl',
            'ovn-sbctl get-connection',
        ]

        schema_dir = '/usr/share/openvswitch'

        nb_tables = self.get_tables_from_schema(os.path.join(
            schema_dir, 'ovn-nb.ovsschema'))
        sb_tables = self.get_tables_from_schema(os.path.join(
            schema_dir, 'ovn-sb.ovsschema'), ['Logical_Flow'])

        self.add_database_output(nb_tables, cmds, 'ovn-nbctl')
        self.add_database_output(sb_tables, cmds, 'ovn-sbctl')

        # If OVN is containerized, we need to run the above commands inside
        # the container.
        if containerized:
            cmds = ['%s exec %s %s' % (self._container_runtime,
                                       self._container_name,
                                       cmd) for cmd in cmds]

        self.add_cmd_output(cmds, foreground=True)

        self.add_copy_spec("/etc/sysconfig/ovn-northd")

        ovs_dbdir = os.environ.get('OVS_DBDIR')
        for dbfile in ['ovnnb_db.db', 'ovnsb_db.db']:
            self.add_copy_spec([
                os.path.join('/var/lib/openvswitch/ovn', dbfile),
                os.path.join('/usr/local/etc/openvswitch', dbfile),
                os.path.join('/etc/openvswitch', dbfile),
                os.path.join('/var/lib/openvswitch', dbfile),
            ])
            if ovs_dbdir:
                self.add_copy_spec(os.path.join(ovs_dbdir, dbfile))

        self.add_journal(units="ovn-northd")


class RedHatOVNCentral(OVNCentral, RedHatPlugin):

    packages = ('openvswitch-ovn-central', 'ovn2.*-central', )


class DebianOVNCentral(OVNCentral, DebianPlugin, UbuntuPlugin):

    packages = ('ovn-central', )



# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin


class HardwareTestSuite(Plugin, RedHatPlugin):
    """Red Hat Hardware Test Suite
    """

    plugin_name = 'hts'
    profiles = ('debug',)

    def setup(self):
        self.add_copy_spec([
            "/etc/httpd/conf.d/hts.conf",
            "/var/hts"
        ])

# vim: set et ts=4 sw=4 :



# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012 Sebastian Wiesner <lunaryorn@googlemail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


"""
    pyudev
    ======

    A binding to libudev_.

    A :class:`Context` object is required for almost any functionality in
    pyudev.  The context provides :class:`Device` enumeration with
    :meth:`Context.list_devices()`.

    Device monitoring is provided by :class:`Monitor` and
    :class:`MonitorObserver`.  With :mod:`pyudev.pyqt4`, :mod:`pyudev.pyside`,
    :mod:`pyudev.glib` and :mod:`pyudev.wx` device monitoring can be integrated
    into the event loop of various GUI toolkits.

    .. _libudev: http://www.kernel.org/pub/linux/utils/kernel/hotplug/libudev/

    .. moduleauthor::  Sebastian Wiesner  <lunaryorn@googlemail.com>
"""

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)


__version__ = '0.15'
__version_info__ = tuple(map(int, __version__.split('.')))
__all__ = ['Context', 'Device']


from pyudev.device import *
from pyudev.core import *
from pyudev.monitor import *



# This file is part of the sos project: https://github.com/sosreport/sos
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from sos.plugins import Plugin, RedHatPlugin, DebianPlugin, UbuntuPlugin


class Auditd(Plugin, RedHatPlugin, DebianPlugin, UbuntuPlugin):
    """Audit daemon information
    """

    plugin_name = 'auditd'
    profiles = ('system', 'security')

    packages = ('audit',)

    def setup(self):
        self.add_copy_spec([
            "/etc/audit/auditd.conf",
            "/etc/audit/audit.rules",
            "/etc/audit/plugins.d/",
            "/etc/audisp/",
        ])
        self.add_cmd_output([
            "ausearch --input-logs -m avc,user_avc -ts today",
            "auditctl -s",
            "auditctl -l"
        ])

        if not self.get_option("all_logs"):
            self.add_copy_spec("/var/log/audit/audit.log")
        else:
            self.add_copy_spec("/var/log/audit")

# vim: set et ts=4 sw=4 :



#! /usr/bin/python
# -*- python -*-
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008, 2009 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
"""
Copyright (c) 2008, 2009  Red Hat Inc.

Abstractions to extract information from the Linux kernel /proc files.
"""
__author__ = "Arnaldo Carvalho de Melo <acme@redhat.com>"
__license__ = "GPLv2 License"

from procfs import *
from sysctl import *
from utilist import *



# Copyright (c) 2016 Hewlett Packard Enterprise Development LP
#
# Redistribution and use of this software in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from requestbuilder import Arg
from requestbuilder.exceptions import ArgumentError

from euca2ools.commands.cloudformation import CloudFormationRequest


class ContinueUpdateRollback(CloudFormationRequest):
    """
    Continue rolling back a stack that has previously failed to roll back
    """

    ARGS = [Arg('StackName', metavar='STACK',
                help='name of the stack to update (required)'),
            Arg('--skip', metavar='RESOURCE', action='append',
                dest='ResourcesToSkip.member', help='''do not re-attempt
                to roll back a given resource, but instead mark it as
                UPDATE_COMPLETE without changing it.  Use this option
                multiple times to skip multiple resources.'''),
            Arg('--role', dest='RoleARN', metavar='ARN', help='''change
                the role the stack uses for this and future operations''')]

    def configure(self):
        CloudFormationRequest.configure(self)
        if (self.args.get('RoleARN') and
                not self.args['RoleARN'].startswith('arn:')):
            raise ArgumentError('argument --role: ARN must begin with "arn:"')
